"""Output SW1L (:mod:`fluidsim.solvers.sw1l.output`)
====================================================

.. autoclass:: OutputBaseSW1L
   :members:
   :private-members:

.. autosummary::
   :toctree:

   print_stdout
   spatial_means
   spect_energy_budget
   spectra
   normal_mode

"""

import numpy as np

from transonic import jit

from fluidsim.base.output import OutputBasePseudoSpectral


@jit
def linear_eigenmode_from_values_1k(
    ux_fft: np.complex128,
    uy_fft: np.complex128,
    eta_fft: np.complex128,
    kx: float,
    ky: float,
    f: "float or int",
    c2: "float or int",
):
    """Compute q, d, a (fft) for a single wavenumber."""
    div_fft = 1j * (kx * ux_fft + ky * uy_fft)
    rot_fft = 1j * (kx * uy_fft - ky * ux_fft)
    q_fft = rot_fft - f * eta_fft
    k2 = kx ** 2 + ky ** 2
    ageo_fft = f * rot_fft / c2 + k2 * eta_fft
    return q_fft, div_fft, ageo_fft


class OutputBaseSW1L(OutputBasePseudoSpectral):
    @staticmethod
    def _complete_info_solver(info_solver):
        """Complete the ParamContainer *info_solver* with child classes to be
        instantiated under *sim.output*.

        This is a static method!
        """
        classes = info_solver.classes.Output._set_child("classes")

        package = "fluidsim.solvers.sw1l.output"

        classes._set_child(
            "PrintStdOut",
            attribs={
                "module_name": package + ".print_stdout",
                "class_name": "PrintStdOutSW1L",
            },
        )

        classes._set_child(
            "PhysFields",
            attribs={
                "module_name": "fluidsim.base.output.phys_fields2d",
                "class_name": "PhysFieldsBase2D",
            },
        )

        classes._set_child(
            "Spectra",
            attribs={
                "module_name": package + ".spectra",
                "class_name": "SpectraSW1LNormalMode",
            },
        )

        classes._set_child(
            "SpatialMeans",
            attribs={
                "module_name": package + ".spatial_means",
                "class_name": "SpatialMeansSW1L",
            },
        )

        attribs = {
            "module_name": package + ".spect_energy_budget",
            "class_name": "SpectralEnergyBudgetSW1L",
        }
        classes._set_child("SpectralEnergyBudget", attribs=attribs)

        attribs = {
            "module_name": package + ".increments",
            "class_name": "IncrementsSW1L",
        }
        classes._set_child("Increments", attribs=attribs)

        attribs = {
            "module_name": "fluidsim.base.output.prob_dens_func",
            "class_name": "ProbaDensityFunc",
        }
        classes._set_child("ProbaDensityFunc", attribs=attribs)

        attribs = {
            "module_name": "fluidsim.base.output.time_signals_fft",
            "class_name": "TimeSignalsK",
        }
        classes._set_child("TimeSignalsK", attribs=attribs)

    @staticmethod
    def _complete_params_with_default(params, info_solver):
        """This static method is used to complete the *params* container."""
        OutputBasePseudoSpectral._complete_params_with_default(
            params, info_solver
        )

        params.output.phys_fields.field_to_plot = "rot"

    def linear_eigenmode_from_values_1k(self, ux_fft, uy_fft, eta_fft, kx, ky):
        """Compute the linear eigenmodes for a single wavenumber."""
        return linear_eigenmode_from_values_1k(
            ux_fft, uy_fft, eta_fft, kx, ky, self.sim.params.f, self.sim.params.c2
        )

    def omega_from_wavenumber(self, k):
        r"""Evaluates the dispersion relation and returns the linear frequency

        .. math:: \omega = \sqrt{f ^ 2 + (ck)^2}

        """
        return np.sqrt(self.sim.params.f ** 2 + self.sim.params.c2 * k ** 2)

    def compute_enstrophy_fft(self):
        r"""Calculate enstrophy from vorticity in the spectral space."""
        rot_fft = self.sim.state.get_var("rot_fft")
        return np.abs(rot_fft) ** 2 / 2.0

    def compute_PV_fft(self):
        r"""Compute Ertel and Charney (QG) potential vorticity.

        .. math:: \zeta_{er} = \frac{f + \zeta}{1 + \eta}
        .. math:: \zeta_{ch} = \zeta - f \eta

        """
        get_var = self.sim.state.get_var
        rot = get_var("rot")
        eta = self.sim.state.state_phys.get_var("eta")
        ErtelPV_fft = self.oper.fft2((self.sim.params.f + rot) / (1.0 + eta))
        rot_fft = get_var("rot_fft")
        eta_fft = get_var("eta_fft")
        CharneyPV_fft = rot_fft - self.sim.params.f * eta_fft
        return ErtelPV_fft, CharneyPV_fft

    def compute_PE_fft(self):
        """Compute Ertel and Charney (QG) potential enstrophy."""
        ErtelPV_fft, CharneyPV_fft = self.compute_PV_fft()
        return (abs(ErtelPV_fft) ** 2 / 2.0, abs(CharneyPV_fft) ** 2 / 2.0)

    def compute_CharneyPE_fft(self):
        """Compute Charney (QG) potential enstrophy."""
        rot_fft = self.sim.state.get_var("rot_fft")
        eta_fft = self.sim.state.get_var("eta_fft")
        CharneyPV_fft = rot_fft - self.sim.params.f * eta_fft
        return abs(CharneyPV_fft) ** 2 / 2.0

    def compute_energies(self):
        """Compute kinetic, available potential and rotational kinetic energies."""
        energyK_fft, energyA_fft, energyKr_fft = self.compute_energies_fft()
        return (
            self.sum_wavenumbers(energyK_fft),
            self.sum_wavenumbers(energyA_fft),
            self.sum_wavenumbers(energyKr_fft),
        )

    def compute_energiesKA(self):
        """Compute K.E. and A.P.E."""
        energyK_fft, energyA_fft = self.compute_energiesKA_fft()
        return (
            self.sum_wavenumbers(energyK_fft),
            self.sum_wavenumbers(energyA_fft),
        )

    def compute_energy(self):
        """Compute total energy by summing K.E. and A.P.E."""
        energyK_fft, energyA_fft = self.compute_energiesKA_fft()
        return self.sum_wavenumbers(energyK_fft) + self.sum_wavenumbers(
            energyA_fft
        )

    def compute_enstrophy(self):
        """Compute total enstrophy."""
        enstrophy_fft = self.compute_enstrophy_fft()
        return self.sum_wavenumbers(enstrophy_fft)

    def compute_lin_energies_fft(self):
        r"""Compute quadratic energies decomposed into contributions from
        potential vorticity (:math:`q`), divergence (:math:`\nabla.\mathbf u`),
        and ageostrophic variable (:math:`a`).

        """
        get_var = self.sim.state.get_var
        ux_fft = get_var("ux_fft")
        uy_fft = get_var("uy_fft")
        eta_fft = get_var("eta_fft")

        q_fft, div_fft, ageo_fft = self.oper.qdafft_from_uxuyetafft(
            ux_fft, uy_fft, eta_fft
        )

        udx_fft, udy_fft = self.oper.vecfft_from_divfft(div_fft)
        energy_dlin_fft = 0.5 * (np.abs(udx_fft) ** 2 + np.abs(udy_fft) ** 2)

        ugx_fft, ugy_fft, etag_fft = self.oper.uxuyetafft_from_qfft(q_fft)
        energy_glin_fft = 0.5 * (
            np.abs(ugx_fft) ** 2
            + np.abs(ugy_fft) ** 2
            + self.sim.params.c2 * np.abs(etag_fft) ** 2
        )

        uax_fft, uay_fft, etaa_fft = self.oper.uxuyetafft_from_afft(ageo_fft)
        energy_alin_fft = 0.5 * (
            np.abs(uax_fft) ** 2
            + np.abs(uay_fft) ** 2
            + self.sim.params.c2 * np.abs(etaa_fft) ** 2
        )

        return energy_glin_fft, energy_dlin_fft, energy_alin_fft


class OutputSW1L(OutputBaseSW1L):
    def compute_energies_fft(self):
        r"""Compute kinetic, available potential and rotational kinetic energies
        in the spectral space.

        .. math:: E_K = (h \mathbf u).\mathbf u / 2
        .. math:: E_A = c^2 \eta^2) / 2
        .. math:: E_{K,r} = (h \mathbf u_r).\mathbf u_r / 2

        """
        get_var = self.sim.state.get_var
        eta_fft = get_var("eta_fft")
        energyA_fft = self.sim.params.c2 * np.abs(eta_fft) ** 2 / 2
        Jx_fft = get_var("Jx_fft")
        Jy_fft = get_var("Jy_fft")
        ux_fft = get_var("ux_fft")
        uy_fft = get_var("uy_fft")
        energyK_fft = (
            np.real(Jx_fft.conj() * ux_fft + Jy_fft.conj() * uy_fft) / 2.0
        )

        rot_fft = get_var("rot_fft")
        uxr_fft, uyr_fft = self.oper.vecfft_from_rotfft(rot_fft)
        rotJ_fft = self.oper.rotfft_from_vecfft(Jx_fft, Jy_fft)
        Jxr_fft, Jyr_fft = self.oper.vecfft_from_rotfft(rotJ_fft)
        energyKr_fft = (
            np.real(Jxr_fft.conj() * uxr_fft + Jyr_fft.conj() * uyr_fft) / 2.0
        )
        return energyK_fft, energyA_fft, energyKr_fft

    def compute_energiesKA_fft(self):
        """Compute K.E. and A.P.E in the spectral space."""
        get_var = self.sim.state.get_var
        eta_fft = get_var("eta_fft")
        energyA_fft = self.sim.params.c2 * np.abs(eta_fft) ** 2 / 2
        Jx_fft = get_var("Jx_fft")
        Jy_fft = get_var("Jy_fft")
        ux_fft = get_var("ux_fft")
        uy_fft = get_var("uy_fft")
        energyK_fft = (
            np.real(Jx_fft.conj() * ux_fft + Jy_fft.conj() * uy_fft) / 2.0
        )

        return energyK_fft, energyA_fft



# -*- coding: utf-8 -*-
""" Anisotropic (:mod:`fluidsim.base.forcing.anisotropic`)
==========================================================

.. autoclass:: TimeCorrelatedRandomPseudoSpectralAnisotropic
   :members:
   :private-members:

"""

from math import degrees

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as patches

from fluidsim.base.forcing.specific import TimeCorrelatedRandomPseudoSpectral


class TimeCorrelatedRandomPseudoSpectralAnisotropic(
    TimeCorrelatedRandomPseudoSpectral
):
    """Random normalized anisotropic forcing.

    .. inheritance-diagram:: TimeCorrelatedRandomPseudoSpectralAnisotropic

    """

    tag = "tcrandom_anisotropic"

    @classmethod
    def _complete_params_with_default(cls, params):
        """This static method is used to complete the *params* container."""
        super(
            TimeCorrelatedRandomPseudoSpectral, cls
        )._complete_params_with_default(params)

        params.forcing._set_child(
            "tcrandom_anisotropic", {"angle": "45°", "kz_negative_enable": False}
        )

    def __init__(self, sim):
        super().__init__(sim)

        if self.params.forcing.normalized.type == "particular_k":
            raise NotImplementedError

    def _compute_cond_no_forcing(self):
        """Computes condition no forcing of the anisotropic case."""
        angle = self.angle

        self.kxmin_forcing = np.sin(angle) * self.kmin_forcing
        self.kxmax_forcing = np.sin(angle) * self.kmax_forcing

        self.kymin_forcing = np.cos(angle) * self.kmin_forcing
        self.kymax_forcing = np.cos(angle) * self.kmax_forcing

        COND_NO_F_KX = np.logical_or(
            self.oper_coarse.KX > self.kxmax_forcing,
            self.oper_coarse.KX < self.kxmin_forcing,
        )

        COND_NO_F_KY = np.logical_or(
            self.oper_coarse.KY > self.kymax_forcing,
            self.oper_coarse.KY < self.kymin_forcing,
        )

        if self.params.forcing.tcrandom_anisotropic.kz_negative_enable:
            COND_NO_F_KY = np.logical_and(
                COND_NO_F_KY,
                np.logical_or(
                    self.oper_coarse.KY < -self.kymax_forcing,
                    self.oper_coarse.KY > -self.kymin_forcing,
                ),
            )

        COND_NO_F = np.logical_or(COND_NO_F_KX, COND_NO_F_KY)
        COND_NO_F[self.oper_coarse.shapeK_loc[0] // 2] = True
        COND_NO_F[:, self.oper_coarse.shapeK_loc[1] - 1] = True
        return COND_NO_F

    # def plot_forcing_1mode_time(self):
    #     """Plots the forcing coherence in time."""

    def plot_forcing_region(self):
        """Plots the forcing region"""
        pforcing = self.params.forcing
        # oper = self.oper

        kxmin_forcing = self.kxmin_forcing
        kxmax_forcing = self.kxmax_forcing
        kymin_forcing = self.kymin_forcing
        kymax_forcing = self.kymax_forcing

        # Define forcing region
        coord_x = kxmin_forcing
        coord_y = kymin_forcing
        width = kxmax_forcing - kxmin_forcing
        height = kymax_forcing - kymin_forcing

        theta1 = 90.0 - degrees(self.angle)
        theta2 = 90.0

        KX = self.oper_coarse.KX
        KY = self.oper_coarse.KY

        fig, ax = plt.subplots()
        ax.set_aspect("equal")

        title = (
            pforcing.type
            + "; "
            + fr"$nk_{{min}} = {pforcing.nkmin_forcing} \delta k_x$; "
            + fr"$nk_{{max}} = {pforcing.nkmax_forcing} \delta k_z$; "
            + r"$\theta = {:.0f}^\circ$; ".format(degrees(self.angle))
            + fr"Forced modes = {self.nb_forced_modes}"
        )

        ax.set_title(title)
        ax.set_xlabel(r"$k_x$")
        ax.set_ylabel(r"$k_z$")

        # Parameters figure
        ax.set_xlim([abs(KX).min(), abs(KX).max()])
        ax.set_ylim([abs(KY).min(), abs(KY).max()])

        # Set ticks 10% of the KX.max and KY.max
        factor = 0.1
        sep_x = abs(KX).max() * factor
        sep_y = abs(KY).max() * factor
        nb_deltakx = int(sep_x // self.oper.deltakx)
        nb_deltaky = int(sep_y // self.oper.deltaky)

        if not nb_deltakx:
            nb_deltakx = 1
        if not nb_deltaky:
            nb_deltaky = 1

        xticks = np.arange(
            abs(KX).min(), abs(KX).max(), nb_deltakx * self.oper.deltakx
        )
        yticks = np.arange(
            abs(KY).min(), abs(KY).max(), nb_deltaky * self.oper.deltaky
        )
        ax.set_xticks(xticks)
        ax.set_yticks(yticks)

        ax.add_patch(
            patches.Rectangle(
                xy=(coord_x, coord_y), width=width, height=height, fill=False
            )
        )

        # width and height arc 50% the length of the axis
        ax.add_patch(
            patches.Arc(
                xy=(0, 0),
                width=abs(KX).max() * 0.5,
                height=abs(KX).max() * 0.5,
                angle=0,
                theta1=theta1,
                theta2=theta2,
            )
        )

        # Plot arc kmin and kmax
        ax.add_patch(
            patches.Arc(
                xy=(0, 0),
                width=2 * self.kmin_forcing,
                height=2 * self.kmin_forcing,
                angle=0,
                theta1=0,
                theta2=90.0,
                linestyle="-.",
            )
        )
        ax.add_patch(
            patches.Arc(
                xy=(0, 0),
                width=2 * self.kmax_forcing,
                height=2 * self.kmax_forcing,
                angle=0,
                theta1=0,
                theta2=90.0,
                linestyle="-.",
            )
        )

        # Plot lines angle & lines forcing region
        ax.plot([0, kxmin_forcing], [0, kymin_forcing], color="k", linewidth=1)
        ax.plot(
            [kxmin_forcing, kxmin_forcing],
            [0, kymin_forcing],
            "k--",
            linewidth=0.8,
        )
        ax.plot(
            [kxmax_forcing, kxmax_forcing],
            [0, kymin_forcing],
            "k--",
            linewidth=0.8,
        )
        ax.plot(
            [0, kxmin_forcing],
            [kymin_forcing, kymin_forcing],
            "k--",
            linewidth=0.8,
        )
        ax.plot(
            [0, kxmin_forcing],
            [kymax_forcing, kymax_forcing],
            "k--",
            linewidth=0.8,
        )

        # Plot forced modes in red
        indices_forcing = np.argwhere(self.COND_NO_F == False)
        for i, index in enumerate(indices_forcing):
            ax.plot(
                KX[0, index[1]],
                KY[index[0], 0],
                "ro",
                label="Forced mode" if i == 0 else "",
            )

        # Location labels 0.8% the length of the axis
        factor = 0.008
        loc_label_y = abs(KY).max() * factor
        loc_label_x = abs(KX).max() * factor

        ax.text(loc_label_x + kxmin_forcing, loc_label_y, r"$k_{x,min}$")
        ax.text(loc_label_x + kxmax_forcing, loc_label_y, r"$k_{x,max}$")
        ax.text(loc_label_x, kymin_forcing + loc_label_y, r"$k_{z,min}$")
        ax.text(loc_label_x, kymax_forcing + loc_label_y, r"$k_{z,max}$")

        # Location label angle \theta
        factor_x = 0.015
        factor_y = 0.15
        loc_label_y = abs(KY).max() * factor_y
        loc_label_x = abs(KX).max() * factor_x

        ax.text(loc_label_x, loc_label_y, r"$\theta$")

        ax.grid(linestyle="--", alpha=0.4)
        ax.legend()



"""Base solver (:mod:`fluidsim.base.solvers.sphericalharmo`)
============================================================

.. autoclass:: InfoSolverSphericalHarmo
   :members:
   :private-members:

.. autoclass:: SimulSphericalHarmo
   :members:
   :private-members:


"""

from ..solvers.pseudo_spect import (
    InfoSolverPseudoSpectral,
    SimulBasePseudoSpectral,
)


class InfoSolverSphericalHarmo(InfoSolverPseudoSpectral):
    """Contain the information on a base pseudo-spectral solver."""

    def _init_root(self):
        """Init. `self` by writting the information on the solver.

        The first-level classes for this base solver are the same as
        for the 2D pseudo-spectral base solver except the class:

        - :class:`fluidsim.operators.operators.OperatorsPseudoSpectral2D`

        """

        super()._init_root()

        here = "fluidsim.base.sphericalharmo"

        self.module_name = here + ".solver"
        self.class_name = "SimulSphericalHarmo"
        self.short_name = "BaseSH"

        self.classes.Operators.module_name = "fluidsim.operators.sphericalharmo"
        self.classes.Operators.class_name = "OperatorsSphericalHarmonics"

        self.classes.State.module_name = here + ".state"
        self.classes.State.class_name = "StateSphericalHarmo"

        self.classes.Output.module_name = here + ".output"
        self.classes.Output.class_name = "Output"


class SimulSphericalHarmo(SimulBasePseudoSpectral):
    """Pseudo-spectral base solver."""

    InfoSolver = InfoSolverSphericalHarmo


Simul = SimulSphericalHarmo

if __name__ == "__main__":

    params = Simul.create_default_params()

    params.short_name_type_run = "test"
    params.time_stepping.USE_CFL = False
    params.time_stepping.t_end = 2.0
    params.time_stepping.deltat0 = 0.1

    sim = Simul(params)
    sim.time_stepping.start()



import unittest

import fluiddyn.util.mpi as mpi
from fluidsim.util.testing import (
    TestSimulConserveOutput,
    classproperty,
    skip_if_no_fluidfft,
)


@skip_if_no_fluidfft
class TestSimulSW1LModified(TestSimulConserveOutput):
    zero = 1e-8

    @classproperty
    def Simul(cls):
        from fluidsim.solvers.sw1l.modified.solver import Simul

        return Simul

    @classmethod
    def init_params(cls):
        super().init_params()
        params = cls.params
        params.output.HAS_TO_SAVE = True
        params.output.ONLINE_PLOT_OK = True
        params.output.spect_energy_budg.HAS_TO_PLOT_SAVED = True

        params.output.periods_save.spatial_means = 0.1
        params.output.periods_save.spect_energy_budg = 0.2

        params.forcing.enable = True
        params.forcing.type = "waves"
        params.forcing.nkmin_forcing = 2
        params.forcing.nkmax_forcing = 4
        params.init_fields.type = "noise"
        params.oper.nx = 16
        params.oper.ny = 8

    def test_spatial_means(self):
        """Test spatial_means."""
        self.sim.output.spatial_means.plot_rates()

    def test_spect_energy_budg(self):
        """Test spect_energy_budg loading and plotting.

        .. TODO::

           Errors

           * if periods_save = 0.25.
             >>> delta_i_plot = int(np.round(delta_t / delta_t_save))
             E ValueError: cannot convert float NaN to integer

        """
        self.get_results("spect_energy_budg")
        self.plot("spect_energy_budg")

    def get_tendencies(self):
        tendencies_fft = self.tendencies_fft
        Fx_fft = tendencies_fft.get_var("ux_fft")
        Fy_fft = tendencies_fft.get_var("uy_fft")
        Feta_fft = tendencies_fft.get_var("eta_fft")

        return Fx_fft, Fy_fft, Feta_fft

    # def test_enstrophy_conservation(self):
    #     """Theoretically not expected to conserve quadratic potential enstrophy"""
    #     pass

    def test_energy_conservation(self):
        """Verify that the energy growth rate due to nonlinear tendencies
        (advection term) must be zero.

        """
        Fx_fft, Fy_fft, Feta_fft = self.get_tendencies()
        state_phys = self.sim.state.state_phys
        ux = state_phys.get_var("ux")
        uy = state_phys.get_var("uy")
        eta = state_phys.get_var("eta")

        oper = self.sim.oper
        Fx = oper.ifft2(Fx_fft)
        Fy = oper.ifft2(Fy_fft)
        Feta = oper.ifft2(Feta_fft)
        A = (ux * Fx + uy * Fy) + self.sim.params.c2 * eta * Feta

        A_fft = oper.fft2(A)
        if mpi.rank == 0:
            self.assertAlmostZero(A_fft[0, 0])

    def test_energy_conservation_fft(self):
        """Verify that the quadratic energy growth rate due to nonlinear
        tendencies (advection term) must be zero in the spectral plane.

        """
        Fx_fft, Fy_fft, Feta_fft = self.get_tendencies()
        state = self.sim.state
        ux_fft = state.get_var("ux_fft")
        uy_fft = state.get_var("uy_fft")
        eta_fft = state.get_var("eta_fft")

        oper = self.sim.oper
        T_ux = (ux_fft.conj() * Fx_fft).real
        T_uy = (uy_fft.conj() * Fy_fft).real
        T_eta = (eta_fft.conj() * Feta_fft).real * self.sim.params.c2
        T_tot = T_ux + T_uy + T_eta
        sum_T = oper.sum_wavenumbers(T_tot)
        self.assertAlmostZero(sum_T)

    def test_energies_fft(self):
        EK_fft, EA_fft, EKr_fft = self.sim.output.compute_energies_fft()

    def test_pv_fft(self):
        ertel_pv_fft, charney_pv_fft = self.sim.output.compute_PV_fft()

    def test_state_compute(self):
        for key in (
            "ux_fft",
            "uy_fft",
            "rot_fft",
            "div_fft",
            "rot",
            "div",
            "q",
            "q_fft",
            "a_fft",
        ):
            var_computed = self.sim.state.compute(key)


if __name__ == "__main__":
    unittest.main()



"""Time stepping (:mod:`fluidsim.base.time_stepping.simple`)
============================================================


Provides:

.. autoclass:: TimeSteppingSimple
   :members:
   :private-members:

"""

import numpy as np

from .base import TimeSteppingBase


class ExactLinearCoefs:
    """Handle the computation of the exact coefficient for the RK4."""

    def __init__(self, time_stepping):
        self.time_stepping = time_stepping
        sim = time_stepping.sim
        self.freq_lin = time_stepping.freq_lin

        self.exact = np.empty_like(self.freq_lin)
        self.exact2 = np.empty_like(self.freq_lin)

        if sim.params.time_stepping.USE_CFL:
            self.get_updated_coefs = self.get_updated_coefs_CLF
            self.dt_old = 0.0
        else:
            self.compute(time_stepping.deltat)
            self.get_updated_coefs = self.get_coefs

    def compute(self, dt):
        """Compute the exact coefficients."""
        f_lin = self.freq_lin
        self.exact = np.exp(-dt * f_lin)
        self.exact2 = np.exp(-dt / 2 * f_lin)
        self.dt_old = dt

    def get_updated_coefs_CLF(self):
        """Get the exact coefficient updated if needed."""
        dt = self.time_stepping.deltat
        if self.dt_old != dt:
            self.compute(dt)
        return self.exact, self.exact2

    def get_coefs(self):
        """Get the exact coefficients as stored."""
        return self.exact, self.exact2


class TimeSteppingSimple(TimeSteppingBase):
    """Time stepping class for pseudo-spectral solvers."""

    def __init__(self, sim):
        super().__init__(sim)

        self._init_freq_lin()
        self._init_compute_time_step()
        self._init_exact_linear_coef()
        self._init_time_scheme()

    def _init_freq_lin(self):
        self.freq_lin = 0.0

    def _init_exact_linear_coef(self):
        self.exact_linear_coefs = ExactLinearCoefs(self)

    def one_time_step_computation(self):
        """One time step"""
        # warning: if the function _time_step_RK comes from an extension, its
        # execution time seems to be attributed to the function
        # one_time_step_computation by cProfile
        self._time_step_RK()
        if np.isnan(np.sum(self.sim.state.state_phys[0])):
            raise ValueError(f"nan at it = {self.it}, t = {self.t:.4f}")

    def _time_step_RK2(self):
        r"""Advance in time with the Runge-Kutta 2 method.

        .. _rk2timescheme:

        Notes
        -----

        .. |p| mathmacro:: \partial

        We consider an equation of the form

        .. math:: \p_t S = \sigma S + N(S),

        The Runge-Kutta 2 method computes an approximation of the
        solution after a time increment :math:`dt`. We denote the
        initial time :math:`t = 0`.

        - Approximation 1:

          .. math:: \p_t \log S = \sigma + \frac{N(S_0)}{S_0},

          Integrating from :math:`t` to :math:`t+dt/2`, it gives:

          .. |SA1halfdt| mathmacro:: S_{A1dt/2}

          .. math:: \SA1halfdt = (S_0 + N_0 dt/2) e^{\frac{\sigma dt}{2}}.


        - Approximation 2:

          .. math::
             \p_t \log S = \sigma
             + \frac{N(\SA1halfdt)}{ \SA1halfdt },

          Integrating from :math:`t` to :math:`t+dt` and retaining
          only the terms in :math:`dt^1` gives:

          .. math::
             S_{dtA2} = S_0 e^{\sigma dt}
             + N(\SA1halfdt) dt e^{\frac{\sigma dt}{2}}.

        """
        dt = self.deltat
        diss, diss2 = self.exact_linear_coefs.get_updated_coefs()

        tendencies_nonlin = self.sim.tendencies_nonlin
        state_phys = self.sim.state.state_phys

        tendencies_n = tendencies_nonlin()
        state_phys_n12 = (state_phys + dt / 2 * tendencies_n) * diss2
        tendencies_phys_n12 = tendencies_nonlin(state_phys_n12)
        self.sim.state.state_phys = (
            state_phys * diss + dt * diss2 * tendencies_phys_n12
        )

    def _time_step_RK4(self):
        r"""Advance in time with the Runge-Kutta 4 method.

        .. _rk4timescheme:

        We consider an equation of the form

        .. math:: \p_t S = \sigma S + N(S),

        The Runge-Kutta 4 method computes an approximation of the
        solution after a time increment :math:`dt`. We denote the
        initial time as :math:`t = 0`. This time scheme uses 4
        approximations. Only the terms in :math:`dt^1` are retained.

        - Approximation 1:

          .. math:: \p_t \log S = \sigma + \frac{N(S_0)}{S_0},

          Integrating from :math:`t` to :math:`t+dt/2` gives:

          .. math:: \SA1halfdt = (S_0 + N_0 dt/2) e^{\sigma \frac{dt}{2}}.

          Integrating from :math:`t` to :math:`t+dt` gives:

          .. math:: S_{A1dt} = (S_0 + N_0 dt) e^{\sigma dt}.


        - Approximation 2:

          .. math::
             \p_t \log S = \sigma
             + \frac{N(\SA1halfdt)}{ \SA1halfdt },

          Integrating from :math:`t` to :math:`t+dt/2` gives:

          .. |SA2halfdt| mathmacro:: S_{A2 dt/2}

          .. math::
             \SA2halfdt = S_0 e^{\sigma \frac{dt}{2}}
             + N(\SA1halfdt) \frac{dt}{2}.

          Integrating from :math:`t` to :math:`t+dt` gives:

          .. math::
             S_{A2dt} = S_0 e^{\sigma dt}
             + N(\SA1halfdt) e^{\sigma \frac{dt}{2}} dt.


        - Approximation 3:

          .. math::
             \p_t \log S = \sigma
             + \frac{N(\SA2halfdt)}{ \SA2halfdt },

          Integrating from :math:`t` to :math:`t+dt` gives:

          .. math::
             S_{A3dt} = S_0 e^{\sigma dt}
             + N(\SA2halfdt) e^{\sigma \frac{dt}{2}} dt.

        - Approximation 4:

          .. math::
             \p_t \log S = \sigma
             + \frac{N(S_{A3dt})}{ S_{A3dt} },

          Integrating from :math:`t` to :math:`t+dt` gives:

          .. math::
             S_{A4dt} = S_0 e^{\sigma dt} + N(S_{A3dt}) dt.


        The final result is a pondered average of the results of 4
        approximations for the time :math:`t+dt`:

          .. math::
             \frac{1}{3} \left[
             \frac{1}{2} S_{A1dt}
             + S_{A2dt} + S_{A3dt}
             + \frac{1}{2} S_{A4dt}
             \right],

        which is equal to:

          .. math::
             S_0 e^{\sigma dt}
             + \frac{dt}{3} \left[
             \frac{1}{2} N(S_0) e^{\sigma dt}
             + N(\SA1halfdt) e^{\sigma \frac{dt}{2}}
             + N(\SA2halfdt) e^{\sigma \frac{dt}{2}}
             + \frac{1}{2} N(S_{A3dt})\right].

        """

        dt = self.deltat
        diss, diss2 = self.exact_linear_coefs.get_updated_coefs()

        tendencies_nonlin = self.sim.tendencies_nonlin
        state_phys = self.sim.state.state_phys

        tendencies_0 = tendencies_nonlin()

        # based on approximation 1
        state_phys_temp = (state_phys + dt / 6 * tendencies_0) * diss
        state_phys_np12_approx1 = (state_phys + dt / 2 * tendencies_0) * diss2

        del tendencies_0
        tendencies_1 = tendencies_nonlin(state_phys_np12_approx1)
        del state_phys_np12_approx1

        # based on approximation 2
        state_phys_temp += dt / 3 * diss2 * tendencies_1
        state_phys_np12_approx2 = state_phys * diss2 + dt / 2 * tendencies_1

        del tendencies_1
        tendencies_2 = tendencies_nonlin(state_phys_np12_approx2)
        del state_phys_np12_approx2

        # based on approximation 3
        state_phys_temp += dt / 3 * diss2 * tendencies_2
        state_phys_np1_approx = state_phys * diss + dt * diss2 * tendencies_2

        del tendencies_2
        tendencies_3 = tendencies_nonlin(state_phys_np1_approx)
        del state_phys_np1_approx

        # result using the 4 approximations
        self.sim.state.state_phys = state_phys_temp + dt / 6 * tendencies_3



"""Time stepping (:mod:`fluidsim.base.time_stepping.base`)
================================================================

Provides:

.. autoclass:: TimeSteppingBase0
   :members:
   :private-members:

.. autoclass:: TimeSteppingBase
   :members:
   :private-members:

"""

from signal import signal
from warnings import warn
from math import pi
from datetime import datetime, timedelta
from time import time

from fluiddyn.util import mpi


def max_abs(arr):
    return max(abs(arr.min()), abs(arr.max()))


class TimeSteppingBase0:
    """Universal time stepping class used for all solvers."""

    @staticmethod
    def _complete_params_with_default(params):
        """This static method is used to complete the *params* container."""
        attribs = {
            "USE_T_END": True,
            "t_end": 10.0,
            "it_end": 10,
            "USE_CFL": False,
            "type_time_scheme": "RK4",
            "deltat0": 0.2,
            "deltat_max": 0.2,
            "cfl_coef": None,
            "max_elapsed": None,
        }
        params._set_child("time_stepping", attribs=attribs)

        params.time_stepping._set_doc(
            """

See :mod:`fluidsim.base.time_stepping.base`.

USE_T_END: bool (default True)

    If True, time step until t > t_end. If False, time step until it >= it_end.

t_end: float

    See documentation USE_T_END.

it_end: int

    If USE_T_END is False, number of time steps.

USE_CFL: bool (default False)

    If True, use an adaptive time step computed in particular with a
    Courant-Friedrichs-Lewy (CFL) condition.

type_time_scheme: str (default "RK4")

    Type of time scheme. Can be in ("RK2", "RK4").

deltat0: float (default 0.2)

    If USE_CFL is False, value of the time step.

deltat_max: float (default 0.2)

    Maximum value of the time step (useful when USE_CFL is True).

cfl_coef: float (default None)

    If not None, clf_coef used in the CFL condition. If None, the value is choosen
    taking into account the time scheme.

max_elapsed: number or str (default None)

    If not None, the computation stops when the elapsed time becomes larger
    than `max_elapsed`. Can be a number (in seconds) or a string (formated as
    "%H:%M:%S").

"""
        )

    def __init__(self, sim):
        self.params = sim.params
        self.sim = sim

        self.it = 0
        self.t = 0

        self._has_to_stop = False

        def handler_signals(signal_number, stack):
            print(f"signal {signal_number} received.")
            self._has_to_stop = True

        try:
            signal(12, handler_signals)
        except ValueError:
            warn("Cannot handle signals - is multithreading on?")

        try:
            param_max_elapsed = self.params.time_stepping.max_elapsed
        except AttributeError:
            # loading an old simulation?
            param_max_elapsed = None

        if param_max_elapsed is not None:
            try:
                self.max_elapsed = float(param_max_elapsed)
            except ValueError:
                t = datetime.strptime(param_max_elapsed, "%H:%M:%S")
                delta_t = timedelta(
                    hours=t.hour, minutes=t.minute, seconds=t.second
                )
                self.max_elapsed = delta_t.total_seconds()

            t_start = None
            if mpi.rank == 0:
                t_start = time()
            if mpi.nb_proc > 1:
                t_start = mpi.comm.bcast(t_start, root=0)
            self._time_should_stop = t_start + self.max_elapsed
        else:
            self.max_elapsed = None

    def start(self):
        """Loop to run the function :func:`one_time_step`.

        If ``self.USE_T_END`` is true, run till ``t >= t_end``,
        otherwise run ``self.it_end`` time steps.
        """
        self.main_loop(print_begin=True, save_init_field=True)
        self.finalize_main_loop()

    def prepare_main_loop(self):
        """Prepare the simulation just before the main loop.

        This function is called automatically in ``main_loop`` if it hasn't
        been called before. It can be used by users for debugging.

        During this preparation, the time of the begining of the simulation is
        set and the outputs are initialized with the initial state.

        """
        self.sim.__enter__()

        output = self.sim.output
        if (
            not hasattr(output, "_has_been_initialized_with_state")
            or not output._has_been_initialized_with_state
        ):
            output.init_with_initialized_state()

        self._prepare_main_loop_called = True

    def finalize_main_loop(self):
        """Finalize the simulation after the main time loop.

        - set the end time
        - finalize the outputs (in particular close the files)
        """
        self.sim.__exit__()

    def main_loop(self, print_begin=False, save_init_field=False):
        """The main time loop!"""

        if not hasattr(self, "_prepare_main_loop_called"):
            self.prepare_main_loop()

        print_stdout = self.sim.output.print_stdout
        if print_begin:
            print_stdout(
                "*************************************\n"
                "Beginning of the computation"
            )
        if save_init_field and self.sim.output._has_to_save:
            self.sim.output.phys_fields.save()

        params_stepping = self.params.time_stepping

        if params_stepping.USE_T_END:
            print_stdout(f"    compute until t = {params_stepping.t_end:10.6g}")
            while self.t < params_stepping.t_end and not self._has_to_stop:
                self.one_time_step()
        else:
            print_stdout(f"    compute until it = {params_stepping.it_end:8d}")
            while self.it < params_stepping.it_end and not self._has_to_stop:
                self.one_time_step()

    def one_time_step(self):
        """Main time stepping function."""
        if self.params.time_stepping.USE_CFL:
            self.compute_time_increment_CLF()
        if self.sim.is_forcing_enabled:
            self.sim.forcing.compute()
        if self.max_elapsed is not None:
            if mpi.rank == 0:
                now = time()
            else:
                now = None
            if mpi.nb_proc > 1:
                now = mpi.comm.bcast(now, root=0)
            if now > self._time_should_stop:
                self.sim.output.print_stdout(
                    "Maximum elapsed time reached. Should stop soon."
                )
                self._has_to_stop = True
        self.sim.output.one_time_step()
        self.one_time_step_computation()
        self.t += self.deltat
        self.it += 1


class TimeSteppingBase(TimeSteppingBase0):
    def _init_compute_time_step(self):

        params_ts = self.params.time_stepping

        if params_ts.USE_CFL:
            if params_ts.cfl_coef is not None:
                self.CFL = params_ts.cfl_coef
            elif any(
                params_ts.type_time_scheme.startswith(scheme)
                for scheme in ["RK2", "Euler"]
            ):
                self.CFL = 0.4
            elif params_ts.type_time_scheme.startswith("RK4"):
                self.CFL = 1.0
            else:
                raise ValueError("Problem name time_scheme")

        else:
            self.deltat = params_ts.deltat0

        self.deltat = params_ts.deltat0

        has_vars = self.sim.state.has_vars

        # TODO: Replace multiple function calls below when has_vars supports
        # `strict` parameter.
        has_ux = has_vars("ux") or has_vars("vx")
        has_uy = has_vars("uy") or has_vars("vy")
        has_uz = has_vars("uz") or has_vars("vz")
        has_eta = has_vars("eta")

        if has_ux and has_uy and has_uz:
            self.compute_time_increment_CLF = (
                self._compute_time_increment_CLF_uxuyuz
            )
        elif has_ux and has_uy and has_eta:
            self.compute_time_increment_CLF = (
                self._compute_time_increment_CLF_uxuyeta
            )
        elif has_ux and has_uy:
            self.compute_time_increment_CLF = (
                self._compute_time_increment_CLF_uxuy
            )
        elif has_ux:
            self.compute_time_increment_CLF = self._compute_time_increment_CLF_ux
        elif hasattr(self.params, "U"):
            self.compute_time_increment_CLF = self._compute_time_increment_CLF_U
        elif params_ts.USE_CFL:
            raise ValueError("params_ts.USE_CFL but no velocity.")

        self.deltat_max = params_ts.deltat_max

    def _init_time_scheme(self):

        params_ts = self.params.time_stepping

        if params_ts.type_time_scheme == "RK2":
            self._time_step_RK = self._time_step_RK2
        elif params_ts.type_time_scheme == "RK4":
            self._time_step_RK = self._time_step_RK4
        else:
            raise ValueError("Problem name time_scheme")

    def is_simul_completed(self):
        """Checks if simulation time or iteration has reached the end according
        to parameters specified.

        """
        if self.params.time_stepping.USE_T_END:
            return self.t >= self.params.time_stepping.t_end

        else:
            return self.it >= self.params.time_stepping.it_end

    def _compute_time_increment_CLF_uxuyuz(self):
        """Compute the time increment deltat with a CLF condition."""
        get_var = self.sim.state.get_var
        ux = get_var("vx")
        uy = get_var("vy")
        uz = get_var("vz")

        if ux.size > 0:
            max_ux = max_abs(ux)
            max_uy = max_abs(uy)
            max_uz = max_abs(uz)
            tmp = (
                max_ux / self.sim.oper.deltax
                + max_uy / self.sim.oper.deltay
                + max_uz / self.sim.oper.deltaz
            )
        else:
            tmp = 0.0

        self._compute_time_increment_CLF_from_tmp(tmp)

    def _compute_time_increment_CLF_from_tmp(self, tmp):

        if mpi.nb_proc > 1:
            tmp = mpi.comm.allreduce(tmp, op=mpi.MPI.MAX)

        if tmp > 0:
            deltat_CFL = self.CFL / tmp
        else:
            deltat_CFL = self.deltat_max

        maybe_new_dt = min(deltat_CFL, self.deltat_max)
        normalize_diff = abs(self.deltat - maybe_new_dt) / maybe_new_dt

        if normalize_diff > 0.02:
            self.deltat = maybe_new_dt

    def _compute_time_increment_CLF_uxuy(self):
        """Compute the time increment deltat with a CLF condition."""

        ux = self.sim.state.get_var("ux")
        uy = self.sim.state.get_var("uy")

        max_ux = max_abs(ux)
        max_uy = max_abs(uy)
        tmp = max_ux / self.sim.oper.deltax + max_uy / self.sim.oper.deltay

        self._compute_time_increment_CLF_from_tmp(tmp)

    def _compute_time_increment_CLF_uxuyeta(self):
        """Compute the time increment deltat with a CLF condition."""

        ux = self.sim.state.get_var("ux")
        uy = self.sim.state.get_var("uy")

        params = self.sim.params
        try:
            f = params.f
        except AttributeError:
            # For spherical solvers, trying to use the dispersion relation for
            # Poincare waves can give absurd phase speeds due to earth radius
            # coming in the relation for wavenumbers kh_l = l * (l + 1) / r
            # f = params.omega
            f = 0

        # Phase speed of the fastest wave from dispersion relation
        if f == 0:
            cph = params.c2 ** 0.5
        else:
            Lh = max(params.oper.Lx, params.oper.Ly)
            k_min = 2 * pi / Lh

            cph = (f ** 2 / k_min ** 2 + params.c2) ** 0.5

        max_ux = max_abs(ux)
        max_uy = max_abs(uy)
        tmp = max_ux / self.sim.oper.deltax + max_uy / self.sim.oper.deltay

        if mpi.nb_proc > 1:
            tmp = mpi.comm.allreduce(tmp, op=mpi.MPI.MAX)

        if tmp > 0:
            deltat_CFL = self.CFL / tmp
        else:
            deltat_CFL = self.deltat_max

        deltat_wave = (
            self.CFL * min(self.sim.oper.deltax, self.sim.oper.deltay) / cph
        )
        maybe_new_dt = min(deltat_CFL, deltat_wave, self.deltat_max)
        normalize_diff = abs(self.deltat - maybe_new_dt) / maybe_new_dt

        if normalize_diff > 0.02:
            self.deltat = maybe_new_dt

    def _compute_time_increment_CLF_ux(self):
        """Compute the time increment deltat with a CLF condition."""
        ux = self.sim.state.get_var("ux")
        max_ux = max_abs(ux)
        tmp = max_ux / self.sim.oper.deltax
        self._compute_time_increment_CLF_from_tmp(tmp)

    def _compute_time_increment_CLF_U(self):
        """Compute the time increment deltat with a CLF condition."""
        max_ux = self.params.U
        tmp = max_ux / self.sim.oper.deltax
        self._compute_time_increment_CLF_from_tmp(tmp)

    def _compute_dispersion_relation(self):
        """Compute time increment from a dispersion relation."""
        pass



"""Base classes for Operators  (:mod:`fluidsim.operators.base`)
===============================================================

Numerical method agnostic base operator classes

Provides:

.. autoclass:: OperatorBase1D
   :members:
   :private-members:

"""

import numpy as np


class OperatorBase:
    def _modify_sim_repr_maker(self, sim_repr_maker):
        if not hasattr(self, "produce_str_describing_oper"):
            return

        repr_oper = self.produce_str_describing_oper()

        if (
            self.axes
            and self.axes != ("lat", "lon")
            and self.params.ONLY_COARSE_OPER
        ):
            str_shape, str_volume = repr_oper.split("_")
            ndim = len(self.axes)
            p_oper = self.params.oper
            if ndim == 1:
                shape_reversed = (p_oper.nx,)
            elif ndim == 2:
                shape_reversed = (p_oper.nx, p_oper.ny)
            elif ndim == 3:
                shape_reversed = (p_oper.nx, p_oper.ny, p_oper.nz)
            else:
                raise NotImplementedError
            str_shape = "x".join(str(n) for n in shape_reversed)
            repr_oper = f"{str_shape}_{str_volume}"

        sim_repr_maker.add_word(repr_oper)

    def _reinit_truncation(self):
        try:
            truncation_shape = self.params.oper.truncation_shape
        except AttributeError:
            pass
        else:
            if truncation_shape == "cubic":
                # nothing to do (default fluidfft)
                pass
            elif truncation_shape == "spherical":
                kmax = self.coef_dealiasing * self.deltakx * self.nx / 2
                self.where_dealiased = np.array(
                    self.K2 >= kmax ** 2, dtype=np.uint8
                )
            elif truncation_shape == "no_multiple_aliases":
                if not hasattr(self, "get_region_multiple_aliases"):
                    raise NotImplementedError
                where_dealiased = self.get_region_multiple_aliases()
                if self.coef_dealiasing:
                    kmax = self.coef_dealiasing * self.deltakx * self.nx / 2
                    where_dealiased |= self.K2 >= kmax ** 2
                self.where_dealiased = np.array(where_dealiased, dtype=np.uint8)
            else:
                raise ValueError(
                    'truncation_shape must be "cubic", "spherical" or "no_multiple_aliases"'
                )


class OperatorsBase1D(OperatorBase):
    @staticmethod
    def _complete_params_with_default(params):
        """This static method is used to complete the *params* container."""
        attribs = {"nx": 48, "Lx": 8.0}
        params._set_child("oper", attribs=attribs)
        return params

    def __init__(self, params=None):
        self.params = params
        self.axes = ("x",)

        self.nx = nx = int(params.oper.nx)
        self.lx = self.Lx = Lx = float(params.oper.Lx)

        self.size = nx
        self.shapeX = self.shapeX_seq = self.shapeX_loc = self.shape = (nx,)
        self.deltax = Lx / nx
        self.x = self.x_seq = self.x_loc = self.xs = self.deltax * np.arange(nx)

    def _str_describing_oper(self):
        if (self.Lx / np.pi).is_integer():
            str_Lx = repr(int(self.Lx / np.pi)) + "pi"
        else:
            str_Lx = f"{self.Lx:.3f}".rstrip("0")

        return str_Lx

    def produce_str_describing_oper(self):
        """Produce a string describing the operator."""
        str_Lx = self._str_describing_oper()
        return ("{}_S" + str_Lx).format(self.nx)

    def produce_long_str_describing_oper(self, oper_method="Base"):
        """Produce a string describing the operator."""
        str_Lx = self._str_describing_oper()
        return (
            f"{oper_method} operator 1D,\n"
            + f"nx = {self.nx:6d}\n"
            + "Lx = "
            + str_Lx
            + "\n"
        )



import unittest

from .test_operators2d import TestCoarse as _TestCoarse


class TestCoarse(_TestCoarse):
    nb_dim = 3

    @property
    def Oper(self):
        from fluidsim.operators.operators3d import OperatorsPseudoSpectral3D

        return OperatorsPseudoSpectral3D


del _TestCoarse

if __name__ == "__main__":
    unittest.main()



"""Solving equations over a sphere (:mod:`fluidsim.solvers.sphere`)
===================================================================

Package containing the source code of the solvers to solve equations over a
sphere.

.. autosummary::
   :toctree:

   ns2d
   sw1l

"""



import unittest
from copy import deepcopy

import numpy as np

from .solver import Simul

from fluiddyn.util import mpi

from fluidsim.util.testing import TestSimul, skip_if_no_fluidfft


skip_if_mpi = unittest.skipIf(
    mpi.nb_proc > 1, "MPI not implemented, for eg. sim.oper.gather_Xspace"
)


@skip_if_no_fluidfft
@skip_if_mpi
class TestSolverSquare1D(TestSimul):
    Simul = Simul

    @classmethod
    def init_params(cls):

        params = cls.params = cls.Simul.create_default_params()

        params.short_name_type_run = "test"
        params.output.sub_directory = "unittests"

        params.oper.nx = 40
        params.oper.Lx = 1.0

        params.nu_2 = 0.01

        params.time_stepping.type_time_scheme = "Euler"
        params.time_stepping.USE_CFL = False
        params.time_stepping.USE_T_END = False
        params.time_stepping.it_end = 4
        params.time_stepping.deltat_max = 0.1

        params.init_fields.type = "gaussian"

        params.output.periods_print.print_stdout = 0.25
        params.output.periods_save.phys_fields = 0.2
        params.output.periods_plot.phys_fields = 0.0
        params.output.phys_fields.field_to_plot = "s"

    def test_simul(self):
        sim = self.sim
        sim.time_stepping.start()
        sim.plot_freq_diss()


@skip_if_no_fluidfft
@skip_if_mpi
class TestTimeStepping(TestSimul):
    Simul = Simul
    deltat = 2e-4
    k_init = 10
    amplitude = 0.7

    @classmethod
    def init_params(cls):

        params = cls.params = cls.Simul.create_default_params()

        params.short_name_type_run = "test"
        params.output.sub_directory = "unittests"

        params.oper.nx = 32
        params.oper.Lx = 2 * np.pi
        params.oper.coef_dealiasing = 0.66

        params.time_stepping.USE_CFL = False
        params.time_stepping.USE_T_END = False
        params.time_stepping.it_end = 0
        params.time_stepping.deltat0 = cls.deltat

        params.init_fields.type = "in_script"

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        sim = cls.sim

        params2 = deepcopy(sim.params)
        params2.oper.nx = 64
        oper2 = type(sim.oper)(params2)

        cls.s_init = 1 + cls.amplitude * np.cos(cls.k_init * sim.oper.x)

        s_init2 = 1 + cls.amplitude * np.cos(cls.k_init * oper2.x)

        s_exact2 = s_init2 / (1 + cls.deltat * s_init2)
        s_exact2_fft = oper2.fft(s_exact2)
        cls.s_exact2_fft = s_exact2_fft[oper2.kx <= sim.oper.kx.max()]

    @classmethod
    def tearDownClass(cls):
        cls.sim.time_stepping.finalize_main_loop()

    def _test_type_time_scheme(
        self,
        type_time_scheme,
        coef_dealiasing=0.66,
        nb_pairs=1,
        nb_steps_compute_new_pair=None,
    ):
        sim = self.sim
        params = sim.params

        params2 = deepcopy(params)
        params2.oper.coef_dealiasing = coef_dealiasing
        sim.oper = type(sim.oper)(params2)

        sim.state.init_statephys_from(s=self.s_init.copy())
        sim.state.statespect_from_statephys()

        params.time_stepping.it_end += 1
        params.time_stepping.type_time_scheme = type_time_scheme
        params.time_stepping.phaseshift_random.nb_pairs = nb_pairs
        params.time_stepping.phaseshift_random.nb_steps_compute_new_pair = (
            nb_steps_compute_new_pair
        )
        sim.time_stepping.init_from_params()
        sim.time_stepping.main_loop()

        s_fft = sim.state.get_var("s_fft")
        assert np.allclose(s_fft, self.s_exact2_fft), abs(
            s_fft - self.s_exact2_fft
        ).max()

    def test_Euler(self):
        self._test_type_time_scheme("Euler")

    def test_Euler_phaseshift(self):
        self._test_type_time_scheme("Euler_phaseshift", 1)

    def test_Euler_phaseshift_random(self):
        self._test_type_time_scheme("Euler_phaseshift_random", 1)

    def test_Euler_phaseshift_random_bis(self):
        self._test_type_time_scheme("Euler_phaseshift_random", 1, 1, 1)

    def test_Euler_phaseshift_random_ter(self):
        self._test_type_time_scheme("Euler_phaseshift_random", 1, 2)

    def test_RK2(self):
        self._test_type_time_scheme("RK2")

    def test_RK2_trapezoid(self):
        self._test_type_time_scheme("RK2_trapezoid")

    def test_RK2_phaseshift(self):
        self._test_type_time_scheme("RK2_phaseshift", 1)

    def test_RK2_phaseshift_random(self):
        self._test_type_time_scheme("RK2_phaseshift_random", 1)

    def test_RK2_phaseshift_exact(self):
        self._test_type_time_scheme("RK2_phaseshift_exact", 1)

    def test_RK4(self):
        self._test_type_time_scheme("RK4")



import datetime

import numpy as np
import h5py
import h5netcdf

from fluiddyn.util import mpi

cfg_h5py = h5py.h5.get_config()

if cfg_h5py.mpi:
    ext = "h5"
    h5pack = h5py
else:
    ext = "nc"
    h5pack = h5netcdf


def _create_variable(group, key, field):
    if ext == "nc":
        if field.ndim == 0:
            dimensions = tuple()
        elif field.ndim == 1:
            dimensions = ("x",)
        elif field.ndim == 2:
            dimensions = ("y", "x")
        elif field.ndim == 3:
            dimensions = ("z", "y", "x")
        try:
            group.create_variable(key, data=field, dimensions=dimensions)
        except AttributeError:
            raise ValueError(
                "Error while creating a netCDF4 variable using group"
                f" of type {type(group)} for key {key}"
            )

    else:
        try:
            group.create_dataset(key, data=field)
        except AttributeError:
            raise ValueError(
                "Error while creating a HDF5 dataset using group"
                f" of type {type(group)} for key {key}"
            )


def save_file(
    path_file,
    state_phys,
    sim_info,
    output_name_run,
    oper,
    time,
    it,
    particular_attr=None,
):
    def create_group_with_attrs(h5file):
        group_state_phys = h5file.create_group("state_phys")
        group_state_phys.attrs["what"] = "obj state_phys for fluidsim"
        group_state_phys.attrs["name_type_variables"] = state_phys.info
        group_state_phys.attrs["time"] = time
        group_state_phys.attrs["it"] = it
        return group_state_phys

    if mpi.nb_proc == 1 or not cfg_h5py.mpi:
        if mpi.rank == 0:
            h5file = h5pack.File(str(path_file), "w")
            group_state_phys = create_group_with_attrs(h5file)
    else:
        h5file = h5pack.File(str(path_file), "w", driver="mpio", comm=mpi.comm)
        group_state_phys = create_group_with_attrs(h5file)

    if mpi.nb_proc == 1:
        for k in state_phys.keys:
            field_seq = state_phys.get_var(k)
            _create_variable(group_state_phys, k, field_seq)
    elif not cfg_h5py.mpi:
        for k in state_phys.keys:
            field_loc = state_phys.get_var(k)
            field_seq = oper.gather_Xspace(field_loc)
            if mpi.rank == 0:
                _create_variable(group_state_phys, k, field_seq)
    else:
        h5file.atomic = False
        ndim = len(oper.shapeX_loc)
        if ndim == 2:
            xstart, ystart = oper.seq_indices_first_X
        elif ndim == 3:
            xstart, ystart, zstart = oper.seq_indices_first_X
        else:
            raise NotImplementedError
        xend = xstart + oper.shapeX_loc[0]
        yend = ystart + oper.shapeX_loc[1]
        for k in state_phys.keys:
            field_loc = state_phys.get_var(k)
            dset = group_state_phys.create_dataset(
                k, oper.shapeX_seq, dtype=field_loc.dtype
            )
            with dset.collective:
                if field_loc.ndim == 2:
                    dset[xstart:xend, ystart:yend] = field_loc
                elif field_loc.ndim == 3:
                    dset[xstart:xend, ystart:yend, :] = field_loc
                else:
                    raise NotImplementedError("Unsupported number of dimensions")
        h5file.close()
        if mpi.rank == 0:
            h5file = h5pack.File(str(path_file), "r+")

    if mpi.rank == 0:
        h5file.attrs["date saving"] = str(datetime.datetime.now()).encode()
        h5file.attrs["name_solver"] = sim_info.solver.short_name
        h5file.attrs["name_run"] = output_name_run
        h5file.attrs["axes"] = np.array(oper.axes, dtype="|S9")
        if particular_attr is not None:
            h5file.attrs["particular_attr"] = particular_attr

        sim_info._save_as_hdf5(hdf5_parent=h5file)
        gp_info = h5file["info_simul"]
        gf_params = gp_info["params"]
        gf_params.attrs["SAVE"] = 1
        gf_params.attrs["NEW_DIR_RESULTS"] = 1
        h5file.close()



"""Initialisation of the fields (:mod:`fluidsim.base.init_fields`)
========================================================================

Provides:

.. autoclass:: InitFieldsBase
   :members:
   :private-members:

"""

import os
from copy import deepcopy

import numpy as np
import h5py
import h5netcdf

from fluiddyn.util import mpi
from fluidsim_core.params import iter_complete_params

from fluidsim.base.setofvariables import SetOfVariables


class InitFieldsBase:
    """Initialization of the fields (base class)."""

    @staticmethod
    def _complete_info_solver(info_solver, classes=None):
        """Static method to complete the ParamContainer info_solver.

        Parameters
        ----------

        info_solver : fluiddyn.util.paramcontainer.ParamContainer

        classes : iterable of classes (SpecificInitFields)

          If a class has the same tag of a default class, it replaces the
          default one (for example with the tag 'noise').

        """
        classesXML = info_solver.classes.InitFields._set_child("classes")

        classes_used = [
            InitFieldsFromFile,
            InitFieldsFromSimul,
            InitFieldsInScript,
            InitFieldsConstant,
            InitFieldsNoise,
        ]

        classes_used = {cls.tag: cls for cls in classes_used}

        if classes is not None:
            for cls in classes:
                classes_used[cls.tag] = cls

        for tag, cls in classes_used.items():
            classesXML._set_child(
                tag,
                attribs={
                    "module_name": cls.__module__,
                    "class_name": cls.__name__,
                },
            )

    @staticmethod
    def _complete_params_with_default(params, info_solver):
        """This static method is used to complete the *params* container."""
        params._set_child(
            "init_fields",
            attribs={
                "type": "constant",
                "available_types": [],
                "modif_after_init": False,
            },
        )

        params.init_fields._set_doc(
            """

See :mod:`fluidsim.base.init_fields`.

type: str (default constant)

    Name of the initialization method.

available_types: list

    Actually not a parameter; just a hint to set `type`.

modif_after_init: bool (default False)

    Used internally when reloading some simulations.

"""
        )

        dict_classes = info_solver.classes.InitFields.import_classes()
        iter_complete_params(params, info_solver, dict_classes.values())

    def __init__(self, sim):

        self.sim = sim
        params = sim.params
        oper = sim.oper

        self.params = params
        self.oper = oper

        type_init = params.init_fields.type
        if type_init not in params.init_fields.available_types:
            raise ValueError(
                type_init + " is not an available flow initialization."
            )

        dict_classes = sim.info.solver.classes.InitFields.import_classes()
        Class = dict_classes[type_init]
        self._specific_init_fields = Class(sim)

    def __call__(self):
        self.sim.state.is_initialized = not bool(
            self.params.init_fields.modif_after_init
        )
        self._specific_init_fields()


class SpecificInitFields:
    tag = "specific"

    @classmethod
    def _complete_params_with_default(cls, params):
        params.init_fields.available_types.append(cls.tag)

    def __init__(self, sim):
        self.sim = sim


class InitFieldsFromFile(SpecificInitFields):

    tag = "from_file"

    @classmethod
    def _complete_params_with_default(cls, params):
        super()._complete_params_with_default(params)
        params.init_fields._set_child(cls.tag, attribs={"path": ""})
        params.init_fields.from_file._set_doc(
            """
path: str
"""
        )

    def __call__(self):

        # Warning: this function is for 2d pseudo-spectral solver!
        # We have to write something more general.

        params = self.sim.params

        path_file = params.init_fields.from_file.path

        if mpi.rank == 0:
            try:
                if os.path.splitext(path_file)[1] == ".nc":
                    h5file = h5netcdf.File(path_file, "r")
                else:
                    h5file = h5py.File(path_file, "r")
            except:
                raise ValueError(
                    "Is file " + path_file + " really a netCDF4/HDF5 file?"
                )

            print("Load state from file:\n[...]" + path_file[-75:])

            try:
                group_oper = h5file["/info_simul/params/oper"]
            except:
                raise ValueError(
                    "The file " + path_file + " does not contain a params object"
                )

            try:
                group_state_phys = h5file["/state_phys"]
            except:
                raise ValueError(
                    "The file "
                    + path_file
                    + " does not contain a state_phys object"
                )

            try:
                axes = h5file.attrs["axes"]
                for r in axes:
                    # for example r can be: 'z', 'y', 'x'
                    r = r.decode("utf-8")
                    nr = f"n{r}"
                    nr_file = group_oper.attrs[nr]
                    if params.oper[nr] != nr_file:
                        raise ValueError(
                            "this is not a correct state for this simulation\n"
                            "self.{0} != params_file.{0}".format(nr)
                        )
                    Lr = f"L{r}"
                    try:
                        Lr_file = group_oper.attrs[Lr]
                    except KeyError:
                        # Length may not be a parameter for eg: sphericalharmo
                        continue
                    else:
                        if params.oper[Lr] != Lr_file:
                            raise ValueError(
                                "this is not a correct state for this simulation\n"
                                "self.params.oper.{0} != params_file.{0}".format(
                                    Lr
                                )
                            )
            except KeyError:
                # Legacy purposes: 2D specific
                nx_file = group_oper.attrs["nx"]
                ny_file = group_oper.attrs["ny"]
                Lx_file = group_oper.attrs["Lx"]
                Ly_file = group_oper.attrs["Ly"]

                if isinstance(nx_file, list):
                    nx_file = nx_file.item()
                    ny_file = ny_file.item()
                    Lx_file = Lx_file.item()
                    Ly_file = Ly_file.item()

                if params.oper.nx != nx_file:
                    raise ValueError(
                        "this is not a correct state for this simulation\n"
                        "self.nx != params_file.nx"
                    )

                if params.oper.ny != ny_file:
                    raise ValueError(
                        "this is not a correct state for this simulation\n"
                        "self.ny != params_file.ny"
                    )

                if params.oper.Lx != Lx_file:
                    raise ValueError(
                        "this is not a correct state for this simulation\n"
                        "self.params.oper.Lx != params_file.Lx"
                    )

                if params.oper.Ly != Ly_file:
                    raise ValueError(
                        "this is not a correct state for this simulation\n"
                        "self.params.oper.Ly != params_file.Ly"
                    )

            keys_state_phys_file = list(group_state_phys.keys())
        else:
            keys_state_phys_file = {}
        if mpi.nb_proc > 1:
            keys_state_phys_file = mpi.comm.bcast(keys_state_phys_file)
        state_phys = self.sim.state.state_phys
        keys_phys_needed = self.sim.info.solver.classes.State.keys_phys_needed
        for k in keys_phys_needed:
            if k in keys_state_phys_file:
                if mpi.rank == 0:
                    field_seq = group_state_phys[k][...]
                else:
                    field_seq = None

                if mpi.nb_proc > 1:
                    field_loc = self.sim.oper.scatter_Xspace(field_seq)
                else:
                    field_loc = field_seq
                state_phys.set_var(k, field_loc)
            else:
                state_phys.set_var(k, self.sim.oper.create_arrayX(value=0.0))
        if mpi.rank == 0:
            time = group_state_phys.attrs["time"]
            try:
                it = group_state_phys.attrs["it"]
            except KeyError:
                # compatibility with older versions
                it = 0
            h5file.close()
        else:
            time = 0.0
            it = 0

        if mpi.nb_proc > 1:
            time = mpi.comm.bcast(time)
            it = mpi.comm.bcast(it)

        if hasattr(self.sim.state, "statespect_from_statephys"):
            self.sim.state.statespect_from_statephys()
            self.sim.state.statephys_from_statespect()
        self.sim.time_stepping.t = time
        self.sim.time_stepping.it = it


def fill_field_fft_2d(field_fft_in, field_fft_out):

    [nk0_seq, nk1_seq] = field_fft_out.shape
    [nk0_seq_in, nk1_seq_in] = field_fft_in.shape

    nk0_min = min(nk0_seq, nk0_seq_in)
    nk1_min = min(nk1_seq, nk1_seq_in)

    # it is a little bit complicate to take into account ky
    for ik1 in range(nk1_min):
        field_fft_out[0, ik1] = field_fft_in[0, ik1]
        field_fft_out[nk0_min // 2, ik1] = field_fft_in[nk0_min // 2, ik1]
    for ik0 in range(1, nk0_min // 2):
        for ik1 in range(nk1_min):
            field_fft_out[ik0, ik1] = field_fft_in[ik0, ik1]
            field_fft_out[-ik0, ik1] = field_fft_in[-ik0, ik1]


def fill_field_fft_3d(field_fft_in, field_fft_out, oper_in, oper_out):

    [nk0, nk1, nk2] = field_fft_out.shape
    [nk0_in, nk1_in, nk2_in] = field_fft_in.shape

    nk0_min = min(nk0, nk0_in)
    nk1_min = min(nk1, nk1_in)
    nk2_min = min(nk2, nk2_in)

    for ik0 in range(nk0_min):
        for ik1 in range(nk1_min):
            for ik2 in range(nk2_min):
                kx_adim, ky_adim, kz_adim = oper_in.kadim_from_ik012rank(
                    ik0, ik1, ik2
                )
                oper_out.set_value_spect(
                    field_fft_out,
                    field_fft_in[ik0, ik1, ik2],
                    kx_adim,
                    ky_adim,
                    kz_adim,
                )


class InitFieldsFromSimul(SpecificInitFields):

    tag = "from_simul"

    def __call__(self):
        self.sim.init_fields.get_state_from_simul = self._get_state_from_simul

    def _make_state_spect_2d(self, sim_in):
        sim = self.sim
        if (
            sim.params.oper.nx == sim_in.params.oper.nx
            and sim.params.oper.ny == sim_in.params.oper.ny
        ):
            return deepcopy(sim_in.state.state_spect)

        # modify resolution
        state_spect = SetOfVariables(like=sim.state.state_spect, value=0.0)
        keys_state_spect = sim_in.info.solver.classes.State["keys_state_spect"]
        for index_key in range(len(keys_state_spect)):
            field_fft_in = sim_in.state.state_spect[index_key]
            field_fft_new_res = state_spect[index_key]
            fill_field_fft_2d(field_fft_in, field_fft_new_res)

        return state_spect

    def _make_state_spect_3d(self, sim_in):
        # sim = self.sim
        sim = self.sim
        oper_in = sim_in.oper

        if (
            sim.params.oper.nx == sim_in.params.oper.nx
            and sim.params.oper.ny == sim_in.params.oper.ny
            and sim.params.oper.nz == sim_in.params.oper.nz
        ):
            return deepcopy(sim_in.state.state_spect)

        # modify resolution
        state_spect = SetOfVariables(like=sim.state.state_spect, value=0.0)
        keys_state_spect = sim_in.info.solver.classes.State["keys_state_spect"]
        for index_key in range(len(keys_state_spect)):
            field_fft_in = sim_in.state.state_spect[index_key]
            field_fft_new_res = state_spect[index_key]
            fill_field_fft_3d(field_fft_in, field_fft_new_res, oper_in, sim.oper)

        return state_spect

    def _get_state_from_simul(self, sim_in):

        # Warning: this function is for 2d pseudo-spectral solver!
        # We have to write something more general.
        # It should be done directly in the operators.

        if mpi.nb_proc > 1:
            raise NotImplementedError(
                "THIS METHOD WON'T BE IMPLEMENTED IN MPI. "
                "The resolution has to be modified in sequential."
            )

        sim = self.sim
        sim.time_stepping.t = sim_in.time_stepping.t

        try:
            sim.params.oper.ny
            try:
                sim.params.oper.nz
            except AttributeError:
                nb_dim = 2
            else:
                sim_in.params.oper.nz
                nb_dim = 3
        except AttributeError:
            nb_dim = 1
        else:
            sim_in.params.oper.ny

        if nb_dim == 1:
            raise NotImplementedError()
        elif nb_dim == 2:
            state_spect = self._make_state_spect_2d(sim_in)
        elif nb_dim == 3:
            state_spect = self._make_state_spect_3d(sim_in)

        if sim.output.name_solver == sim_in.output.name_solver:
            sim.state.state_spect = state_spect
        else:  # complicated case... untested solution !
            raise NotImplementedError
            # state_spect = SetOfVariables('state_spect')
            # for k in sim.info.solver.classes.State["keys_state_spect"]:
            #     if k in sim_in.info.solver.classes.State["keys_state_spect"]:
            #         sim.state.state_spect[k] = state_spect[k]
            #     else:
            #         sim.state.state_spect[k] = self.oper.create_arrayK(value=0.0)

        sim.state.statephys_from_statespect()


class InitFieldsInScript(SpecificInitFields):

    tag = "in_script"

    def __call__(self):
        self.sim.state.is_initialized = False
        self.sim.output.print_stdout(
            "Manual initialization of the fields is selected. "
            "Do not forget to initialize them."
        )


class InitFieldsConstant(SpecificInitFields):

    tag = "constant"

    @classmethod
    def _complete_params_with_default(cls, params):
        super()._complete_params_with_default(params)
        params.init_fields._set_child(cls.tag, attribs={"value": 1.0})
        params.init_fields.constant._set_doc(
            """
value: float (default 1.)
"""
        )

    def __call__(self):
        value = self.sim.params.init_fields.constant.value
        self.sim.state.state_phys.initialize(value)

        if hasattr(self.sim.state, "statespect_from_statephys"):
            self.sim.state.statespect_from_statephys()


class InitFieldsNoise(SpecificInitFields):
    """Initialize the state with noise."""

    tag = "noise"

    @classmethod
    def _complete_params_with_default(cls, params):
        """Complete the `params` container (class method)."""
        super()._complete_params_with_default(params)
        params.init_fields._set_child(cls.tag, attribs={"max": 1.0})
        params.init_fields.noise._set_doc(
            """
max: float (default 1.)
"""
        )

    def __call__(self):
        state_phys = self.sim.state.state_phys
        state_phys[...] = (
            self.sim.params.init_fields.noise.max
            / 0.5
            * (np.random.rand(*state_phys.shape) - 0.5)
        )

        if hasattr(self.sim.state, "statespect_from_statephys"):
            self.sim.state.statespect_from_statephys()
            self.sim.state.statephys_from_statespect()



"""Operators 3d (:mod:`fluidsim.operators.operators3d`)
=======================================================

Provides

.. autoclass:: OperatorsPseudoSpectral3D
   :members:
   :private-members:

"""

from math import pi
from copy import deepcopy
from random import uniform

import numpy as np

from transonic import boost, Array, Transonic
from fluiddyn.util import mpi
from fluiddyn.util.mpi import nb_proc, rank
from fluidfft.fft3d.operators import OperatorsPseudoSpectral3D as _Operators

from fluidsim.base.setofvariables import SetOfVariables
from fluidsim.base.params import Parameters

from .operators2d import OperatorsPseudoSpectral2D as OpPseudoSpectral2D
from .. import _is_testing
from .base import OperatorBase

ts = Transonic()

Asov = Array[np.complex128, "4d"]
Aui8 = Array[np.uint8, "3d"]
Ac = Array[np.complex128, "3d"]
Af = Array[np.float64, "3d"]


@boost
def dealiasing_setofvar(sov: Asov, where_dealiased: Aui8):
    """Dealiasing 3d setofvar object.

    Parameters
    ----------

    sov : 4d ndarray
        A set of variables array.

    where_dealiased : 3d ndarray
        A 3d array of "booleans" (actually uint8).

    """
    nk, n0, n1, n2 = sov.shape

    for i0 in range(n0):
        for i1 in range(n1):
            for i2 in range(n2):
                if where_dealiased[i0, i1, i2]:
                    for ik in range(nk):
                        sov[ik, i0, i1, i2] = 0.0


@boost
def dealiasing_variable(ff_fft: Ac, where_dealiased: Aui8):
    """Dealiasing 3d array"""
    n0, n1, n2 = ff_fft.shape

    for i0 in range(n0):
        for i1 in range(n1):
            for i2 in range(n2):
                if where_dealiased[i0, i1, i2]:
                    ff_fft[i0, i1, i2] = 0.0


def dealiasing_setofvar_numpy(sov: Asov, where_dealiased: Aui8):
    for i in range(sov.shape[0]):
        sov[i][np.nonzero(where_dealiased)] = 0.0


def dealiasing_variable_numpy(ff_fft: Ac, where_dealiased: Aui8):
    ff_fft[np.nonzero(where_dealiased)] = 0.0


@boost
def compute_energy_from_1field(arr: Ac):
    return 0.5 * np.abs(arr) ** 2


@boost
def compute_energy_from_1field_with_coef(arr: Ac, coef: float):
    return 0.5 * coef * np.abs(arr) ** 2


@boost
def compute_energy_from_2fields(vx: Ac, vy: Ac):
    return 0.5 * (np.abs(vx) ** 2 + np.abs(vy) ** 2)


@boost
def compute_energy_from_3fields(
    vx: Ac,
    vy: Ac,
    vz: Ac,
):
    return 0.5 * (np.abs(vx) ** 2 + np.abs(vy) ** 2 + np.abs(vz) ** 2)


if not ts.is_transpiling and not ts.is_compiled and not _is_testing:
    # for example if Pythran is not available
    dealiasing_variable = dealiasing_variable_numpy
    dealiasing_setofvar = dealiasing_setofvar_numpy
elif ts.is_transpiling:
    _Operators = object


if nb_proc > 1:
    MPI = mpi.MPI
    comm = mpi.comm


@boost
class OperatorsPseudoSpectral3D(_Operators, OperatorBase):
    """Provides fast Fourier transform functions and 3D operators.

    Uses fft operators that implement the methods:

    - ifft
    - fft
    - get_shapeX_loc
    - get_shapeX_seq
    - get_shapeK_loc
    - get_shapeK_seq
    - get_dimX_K
    - get_seq_indices_first_K

    - get_k_adim_loc
    - sum_wavenumbers
    - build_invariant_arrayK_from_2d_indices12X

    """

    Kx: Af
    Ky: Af
    Kz: Af
    inv_K_square_nozero: Af
    inv_Kh_square_nozero: Af
    deltax: float
    deltay: float
    deltaz: float

    @classmethod
    def _create_default_params(cls):
        params = Parameters(tag="params", attribs={"ONLY_COARSE_OPER": False})
        cls._complete_params_with_default(params)
        return params

    @staticmethod
    def _complete_params_with_default(params):
        """This static method is used to complete the *params* container."""
        attribs = {
            "type_fft": "default",
            "type_fft2d": "sequential",
            "coef_dealiasing": 2.0 / 3,
            "nx": 48,
            "ny": 48,
            "nz": 48,
            "Lx": 2 * pi,
            "Ly": 2 * pi,
            "Lz": 2 * pi,
            "truncation_shape": "cubic",
            "NO_SHEAR_MODES": False,
        }
        params._set_child("oper", attribs=attribs)
        params.oper._set_doc(
            """

See the `documentation of fluidfft <http://fluidfft.readthedocs.io>`_ (in
particular of the `3d operator class
<http://fluidfft.readthedocs.io/en/latest/generated/fluidfft.fft3d.operators.html>`_).

type_fft: str

    Method for the FFT (as defined by fluidfft).

type_fft2d: str

    Method for the 2d FFT.

coef_dealiasing: float

    dealiasing coefficient.

nx: int

    Number of points over the x-axis (last dimension in the physical space).

ny: int

    Number of points over the y-axis (second dimension in the physical space).

nz: int

    Number of points over the z-axis (first dimension in the physical space).

Lx, Ly and Lz: float

    Length of the edges of the numerical domain.

"""
        )

    def __init__(self, params=None):

        self.params = params
        self.axes = ("z", "y", "x")

        params.oper.nx = int(params.oper.nx)
        params.oper.ny = int(params.oper.ny)
        params.oper.nz = int(params.oper.nz)

        if params.ONLY_COARSE_OPER:
            nx = ny = nz = 4
        else:
            nx = params.oper.nx
            ny = params.oper.ny
            nz = params.oper.nz

        super().__init__(
            nx,
            ny,
            nz,
            params.oper.Lx,
            params.oper.Ly,
            params.oper.Lz,
            fft=params.oper.type_fft,
            coef_dealiasing=params.oper.coef_dealiasing,
        )

        # compatibility for fluidfft <= 0.3.0
        if not hasattr(self, "oper_fft") and hasattr(self, "_op_fft"):
            self.oper_fft = self._op_fft

        # problem here type_fft
        params2d = deepcopy(params)
        params2d.oper.type_fft = params2d.oper.type_fft2d
        fft = params2d.oper.type_fft

        if (
            any([fft.startswith(s) for s in ["fluidfft.fft2d.", "fft2d."]])
            or fft in ("default", "sequential")
            or fft is None
        ):
            self.oper2d = OpPseudoSpectral2D(params2d)
        else:
            raise ValueError

        self.ifft2 = self.ifft2d = self.oper2d.ifft2
        self.fft2 = self.fft2d = self.oper2d.fft2
        if not self.is_sequential:
            self.iK0loc_start = self.seq_indices_first_K[0]
            self.nk0_loc, self.nk1_loc, self.nk2_loc = self.shapeK_loc

            self.iK0loc_start_rank = np.array(comm.allgather(self.iK0loc_start))
            nk2_loc = self.shapeK_loc[2]
            nk2_loc_rank = np.array(comm.allgather(nk2_loc))
            a = nk2_loc_rank
            self.SAME_SIZE_IN_ALL_PROC = (a >= a.max()).all()
            self.dimX_K = self.oper_fft.get_dimX_K()
        else:
            self.SAME_SIZE_IN_ALL_PROC = True

        self._reinit_truncation()

        try:
            NO_SHEAR_MODES = self.params.oper.NO_SHEAR_MODES
        except AttributeError:
            pass
        else:
            if NO_SHEAR_MODES:
                COND_NOSHEAR = self.Kx ** 2 + self.Ky ** 2 == 0.0
                self.where_dealiased = np.array(
                    np.logical_or(COND_NOSHEAR, self.where_dealiased),
                    dtype=np.uint8,
                )

    def get_region_multiple_aliases(self):
        aliases_x = abs(self.Kx) >= 2 / 3 * self.deltakx * self.nx / 2
        aliases_y = abs(self.Ky) >= 2 / 3 * self.deltaky * self.ny / 2
        aliases_z = abs(self.Kz) >= 2 / 3 * self.deltakz * self.nz / 2
        return (
            (aliases_x & aliases_y)
            | (aliases_y & aliases_z)
            | (aliases_z & aliases_x)
        )

    @property
    def K2_not0(self):
        K2_not0 = np.copy(self.K2)
        if sum(self.seq_indices_first_K) == 0:
            K2_not0[0, 0, 0] = 1e-14
        return K2_not0

    @property
    def K4(self):
        return self.K2 ** 2

    def build_invariant_arrayX_from_2d_indices12X(self, arr2d, oper2d=None):
        """Build a 3D array from a 2D array"""
        if oper2d is None:
            oper2d = self.oper2d

        if mpi.nb_proc == 1:
            return self.oper_fft.build_invariant_arrayX_from_2d_indices12X(
                oper2d.oper_fft, arr2d
            )

        if mpi.rank > 0:
            assert arr2d is None
        arr2d = mpi.comm.bcast(arr2d, root=0)

        n0_loc, n1_loc, n2_loc = self.shapeX_loc

        if mpi.rank == 0:
            shapeX_loc_2d = oper2d.shapeX_loc
        else:
            shapeX_loc_2d = None
        shapeX_loc_2d = mpi.comm.bcast(shapeX_loc_2d, root=0)

        if shapeX_loc_2d != (n1_loc, n2_loc):
            raise NotImplementedError
            # ind0seq_first, ind1seq_first, ind2seq_first = \
            #     self.oper_fft.get_seq_indices_first_K()

        return np.stack(n0_loc * [arr2d])

    def build_invariant_arrayK_from_2d_indices12X(self, arr2d):
        """Build a 3D array from a 2D array"""
        return self.oper_fft.build_invariant_arrayK_from_2d_indices12X(
            self.oper2d, arr2d
        )

    def dealiasing(self, *args):
        """Dealiasing of SetOfVariables or np.ndarray"""
        for thing in args:
            if isinstance(thing, SetOfVariables):
                dealiasing_setofvar(thing, self.where_dealiased)
            elif isinstance(thing, np.ndarray):
                dealiasing_variable(thing, self.where_dealiased)

    def put_coarse_array_in_array_fft(
        self, arr_coarse, arr, oper_coarse, shapeK_loc_coarse
    ):
        """Put the values contained in a coarse array in an array.

        Both arrays are in Fourier space.

        """
        if arr.ndim == 4:
            if rank == 0:
                if arr_coarse.ndim != 4:
                    raise ValueError

            for ikey in range(arr.shape[0]):
                if rank == 0:
                    arr3d_coarse = arr_coarse[ikey]
                else:
                    arr3d_coarse = None
                self.put_coarse_array_in_array_fft(
                    arr3d_coarse, arr[ikey], oper_coarse, shapeK_loc_coarse
                )
            return

        nkzc, nkyc, nkxc = shapeK_loc_coarse

        if nb_proc > 1:
            nk0, nk1, nk2 = self.shapeK_seq
            if mpi.rank == 0:
                if self.dimX_K == (1, 0, 2):
                    fck_fft = np.zeros((nkyc, nkzc, nkxc), dtype=np.complex128)
                    for i0 in range(nkzc):
                        for i1 in range(nkyc):
                            fck_fft[i1, i0, :] = arr_coarse[i0, i1, :]
                if self.dimX_K == (2, 1, 0):
                    fck_fft = np.zeros((nkxc, nkyc, nkzc), dtype=np.complex128)
                    for i0 in range(nkzc):
                        for i1 in range(nkyc):
                            for i2 in range(nkxc):
                                fck_fft[i2, i1, i0] = arr_coarse[i0, i1, i2]

            if self.dimX_K == (1, 0, 2):
                nk1c, nk0c, nk2c = shapeK_loc_coarse
            elif self.dimX_K == (2, 1, 0):
                nk2c, nk1c, nk0c = shapeK_loc_coarse

            for ik0c in range(nk0c):
                ik1c = 0
                ik2c = 0
                ik0 = _ik_from_ikc(ik0c, nk0c, nk0)
                rank_ik, ik0loc, ik1loc, ik2loc = self.where_is_wavenumber(
                    ik0, ik1c, ik2c
                )
                if mpi.rank == 0:
                    fc1D = fck_fft[ik0c, :, :]
                if rank_ik != 0:
                    # message fc1D
                    if mpi.rank == rank_ik:
                        fc1D = np.empty([nk1c, nk2c], dtype=np.complex128)
                    if mpi.rank == 0 or mpi.rank == rank_ik:
                        fc1D = np.ascontiguousarray(fc1D)
                    if mpi.rank == 0:
                        mpi.comm.Send(
                            [fc1D, mpi.MPI.COMPLEX], dest=rank_ik, tag=ik0c
                        )
                    elif mpi.rank == rank_ik:
                        mpi.comm.Recv([fc1D, mpi.MPI.COMPLEX], source=0, tag=ik0c)
                if mpi.rank == rank_ik:
                    # copy
                    for ik1c in range(nk1c):
                        ik1 = _ik_from_ikc(ik1c, nk1c, nk1)
                        arr[ik0loc, ik1, 0:nk2c] = fc1D[ik1c, :]

        else:
            nkz, nky, nkx = self.shapeK_seq
            for ikzc in range(nkzc):
                ikz = _ik_from_ikc(ikzc, nkzc, nkz)
                for ikyc in range(nkyc):
                    iky = _ik_from_ikc(ikyc, nkyc, nky)
                    for ikxc in range(nkxc):
                        arr[ikz, iky, ikxc] = arr_coarse[ikzc, ikyc, ikxc]

    def coarse_seq_from_fft_loc(self, f_fft, shapeK_loc_coarse):
        """Return a coarse field in K space."""
        nkzc, nkyc, nkxc = shapeK_loc_coarse
        if nb_proc > 1:
            if self.shapeK_seq[1:2] != self.shapeK_loc[1:2]:
                raise NotImplementedError()

            if self.dimX_K == (1, 0, 2):
                nk1c, nk0c, nk2c = shapeK_loc_coarse
            elif self.dimX_K == (2, 1, 0):
                nk2c, nk1c, nk0c = shapeK_loc_coarse

            fc_fft_tmp = np.empty([nk0c, nk1c, nk2c], np.complex128)
            nk0, nk1, nk2 = self.shapeK_seq
            f1d_temp = np.empty([nk1c, nk2c], np.complex128)

            for ik0c in range(nk0c):
                ik1c = 0
                ik2c = 0
                ik0 = _ik_from_ikc(ik0c, nk0c, nk0)
                rank_ik, ik0loc, ik1loc, ik1loc = self.where_is_wavenumber(
                    ik0, ik1c, ik2c
                )
                if rank == rank_ik:
                    # create f1d_temp
                    for ik1c in range(nk1c):
                        ik1 = _ik_from_ikc(ik1c, nk1c, nk1)
                        f1d_temp[ik1c, :] = f_fft[ik0loc, ik1, 0:nk2c]

                if rank_ik != 0:
                    # message f1d_temp
                    if rank == 0:
                        comm.Recv(
                            [f1d_temp, MPI.DOUBLE_COMPLEX],
                            source=rank_ik,
                            tag=ik0c,
                        )
                    elif rank == rank_ik:
                        comm.Send(
                            [f1d_temp, MPI.DOUBLE_COMPLEX], dest=0, tag=ik0c
                        )
                if rank == 0:
                    # copy into fc_fft
                    fc_fft_tmp[ik0c] = f1d_temp.copy()
            fc_fft = np.zeros(shapeK_loc_coarse, dtype=np.complex128)
            if rank == 0:
                if self.dimX_K == (1, 0, 2):
                    for i0 in range(nkzc):
                        for i1 in range(nkyc):
                            fc_fft[i0, i1, :] = fc_fft_tmp[i1, i0, :]
                if self.dimX_K == (2, 1, 0):
                    for i0 in range(nkzc):
                        for i1 in range(nkyc):
                            for i2 in range(nkxc):
                                fc_fft[i0, i1, i2] = fc_fft_tmp[i2, i1, i0]
        else:
            fc_fft = np.empty(shapeK_loc_coarse, np.complex128)
            nkz, nky, nkx = self.shapeK_seq
            for ikzc in range(nkzc):
                ikz = _ik_from_ikc(ikzc, nkzc, nkz)
                for ikyc in range(nkyc):
                    iky = _ik_from_ikc(ikyc, nkyc, nky)
                    for ikxc in range(nkxc):
                        fc_fft[ikzc, ikyc, ikxc] = f_fft[ikz, iky, ikxc]
        return fc_fft

    def where_is_wavenumber(self, ik0, ik1, ik2):
        """Give local indices and rank from the sequential indices"""
        nk0_seq, nk1_seq, nk2_seq = self.shapeK_seq

        if ik0 >= nk0_seq:
            raise ValueError("not good :-) ik0_seq >= nk0_seq")

        if nb_proc == 1:
            rank_k = 0
            ik0_loc = ik0
        else:
            if self.SAME_SIZE_IN_ALL_PROC:
                rank_k = int(np.floor(float(ik0) / self.nk0_loc))
                if ik0 >= self.nk0_loc * nb_proc:
                    raise ValueError(
                        "not good :-) ik0_seq >= self.nk0_loc * mpi.nb_proc\n"
                        "ik0_seq, self.nk0_loc, mpi.nb_proc = "
                        f"{ik0}, {self.nk0_loc}, {nb_proc}"
                    )

            else:
                rank_k = 0
                while rank_k < self.nb_proc - 1 and (
                    not (
                        self.iK0loc_start_rank[rank_k] <= ik0
                        and ik0 < self.iK0loc_start_rank[rank_k + 1]
                    )
                ):
                    rank_k += 1

            ik0_loc = ik0 - self.iK0loc_start_rank[rank_k]

        ik1_loc = ik1
        if ik1_loc < 0:
            ik1_loc = self.nk1_loc + ik1_loc

        ik2_loc = ik2
        if ik2_loc < 0:
            ik2_loc = self.nk2_loc + ik2_loc

        return rank_k, ik0_loc, ik1_loc, ik2_loc

    @boost
    def urudfft_from_vxvyfft(self, vx_fft: Ac, vy_fft: Ac):
        """Compute toroidal and poloidal horizontal velocities.

        urx_fft, ury_fft contain shear modes!

        """
        inv_Kh_square_nozero = self.Kx ** 2 + self.Ky ** 2
        inv_Kh_square_nozero[inv_Kh_square_nozero == 0] = 1e-14
        inv_Kh_square_nozero = 1 / inv_Kh_square_nozero

        kdotu_fft = self.Kx * vx_fft + self.Ky * vy_fft
        udx_fft = kdotu_fft * self.Kx * inv_Kh_square_nozero
        udy_fft = kdotu_fft * self.Ky * inv_Kh_square_nozero

        urx_fft = vx_fft - udx_fft
        ury_fft = vy_fft - udy_fft

        return urx_fft, ury_fft, udx_fft, udy_fft

    @boost
    def divhfft_from_vxvyfft(self, vx_fft: Ac, vy_fft: Ac):
        """Compute the horizontal divergence in spectral space."""
        return 1j * (self.Kx * vx_fft + self.Ky * vy_fft)

    @boost
    def vxvyfft_from_rotzfft(self, rotz_fft: Ac):

        inv_Kh_square_nozero = self.Kx ** 2 + self.Ky ** 2
        inv_Kh_square_nozero[inv_Kh_square_nozero == 0] = 1e-14
        inv_Kh_square_nozero = 1 / inv_Kh_square_nozero

        vx_fft = 1j * self.Ky * inv_Kh_square_nozero * rotz_fft
        vy_fft = -1j * self.Kx * inv_Kh_square_nozero * rotz_fft
        return vx_fft, vy_fft

    def get_grid1d_seq(self, axe="x"):

        if axe not in ("x", "y", "z"):
            raise ValueError

        if self.params.ONLY_COARSE_OPER:
            number_points = getattr(self.params.oper, "n" + axe)
            length = getattr(self, "L" + axe)
            return np.linspace(0, length, number_points)
        else:
            return getattr(self, axe + "_seq")

    def project_fft_on_realX(self, f_fft):
        return self.fft(self.ifft(f_fft))

    def _ikxyzseq_from_ik012rank(self, ik0, ik1, ik2, rank=0):
        """Give the sequential indices (xyz) from the local indices and the rank"""
        if self._is_mpi_lib:
            # much more complicated in this case
            raise NotImplementedError
        dimX_K = self.oper_fft.get_dimX_K()
        if dimX_K == (0, 1, 2):
            ikz, iky, ikx = ik0, ik1, ik2
        else:
            raise NotImplementedError(
                f"dimX_K={dimX_K} not implemented ({self.oper_fft.__class__})"
            )
        return ikx, iky, ikz

    def _ik012rank_from_ikxyzseq(self, ikx, iky, ikz):
        """Give the local indices and the rank from "sequential" indices (xyz)"""
        if self._is_mpi_lib:
            # much more complicated in this case
            raise NotImplementedError
        rank_k = 0
        dimX_K = self.oper_fft.get_dimX_K()
        if dimX_K == (0, 1, 2):
            ik0, ik1, ik2 = ikz, iky, ikx
        else:
            raise NotImplementedError
        return ik0, ik1, ik2, rank_k

    def _kadim_from_ikxyzseq(self, ikx, iky, ikz):
        """Give the adimensional wavenumbers from sequential indices"""
        kx_adim = ikx
        ky_adim = _kadim_from_ik(iky, self.ny)
        kz_adim = _kadim_from_ik(ikz, self.nz)
        return kx_adim, ky_adim, kz_adim

    def _ikxyzseq_from_kadim(self, kx_adim, ky_adim, kz_adim):
        """Give the sequential indices from the adimensional wavenumbers"""
        ikx = kx_adim
        iky = _ik_from_kadim(ky_adim, self.ny)
        ikz = _ik_from_kadim(kz_adim, self.nz)
        return ikx, iky, ikz

    def kadim_from_ik012rank(self, ik0, ik1, ik2, rank=0):
        """Give the adimensional wavenumbers from local indices and rank"""
        ikx, iky, ikz = self._ikxyzseq_from_ik012rank(ik0, ik1, ik2, rank)
        return self._kadim_from_ikxyzseq(ikx, iky, ikz)

    def ik012rank_from_kadim(self, kx_adim, ky_adim, kz_adim):
        """Give the local indices and rank from the adimensional wavenumbers"""
        ikx, iky, ikz = self._ikxyzseq_from_kadim(kx_adim, ky_adim, kz_adim)
        return self._ik012rank_from_ikxyzseq(ikx, iky, ikz)

    def set_value_spect(
        self, arr_fft, value, kx_adim, ky_adim, kz_adim, from_rank=0
    ):
        """Set a value in a spectral array given the adimensional wavenumber"""
        ik0, ik1, ik2, rank_k = self.ik012rank_from_kadim(
            kx_adim, ky_adim, kz_adim
        )
        if mpi.rank != rank_k or from_rank != 0:
            raise NotImplementedError
        # print("-" * 20)
        # print(f"ik0, ik1, ik2             = ({ik0:4d}, {ik1:4d}, {ik2:4d})")
        arr_fft[ik0, ik1, ik2] = value

    def get_value_spect(self, arr_fft, kx_adim, ky_adim, kz_adim, to_rank=0):
        """Get a value in a spectral array given the adimensional wavenumber"""
        ik0, ik1, ik2, rank_k = self.ik012rank_from_kadim(
            kx_adim, ky_adim, kz_adim
        )
        if mpi.rank != rank_k or to_rank != 0:
            raise NotImplementedError
        return arr_fft[ik0, ik1, ik2]

    @boost
    def get_phases_random(self):
        # Not supported by Pythran 0.9.5!
        # alpha_x, alpha_y, alpha_z = np.random.uniform(-0.5, 0.5, 3)
        alpha_x, alpha_y, alpha_z = tuple(uniform(-0.5, 0.5) for _ in range(3))
        beta_x = alpha_x + 0.5 if alpha_x < 0 else alpha_x - 0.5
        beta_y = alpha_y + 0.5 if alpha_y < 0 else alpha_y - 0.5
        beta_z = alpha_z + 0.5 if alpha_z < 0 else alpha_z - 0.5

        phase_alpha = (
            alpha_x * self.deltax * self.Kx
            + alpha_y * self.deltay * self.Ky
            + alpha_z * self.deltaz * self.Kz
        )
        phase_beta = (
            beta_x * self.deltax * self.Kx
            + beta_y * self.deltay * self.Ky
            + beta_z * self.deltaz * self.Kz
        )
        return phase_alpha, phase_beta


def _ik_from_ikc(ikc, nkc, nk):
    if ikc <= nkc / 2.0:
        ik = ikc
    else:
        knodim = ikc - nkc
        ik = knodim + nk
    return ik


def _kadim_from_ik(ik, nk, first=False):
    if first or ik <= nk // 2:
        return ik
    return ik - nk


def _ik_from_kadim(k_adim, nk, first=False):
    if first or k_adim >= 0:
        return k_adim
    return nk + k_adim


if __name__ == "__main__":
    n = 4

    p = OperatorsPseudoSpectral3D._create_default_params()

    p.oper.nx = n
    p.oper.ny = 2 * n
    p.oper.nz = 4 * n

    # p.oper.type_fft = 'fftwpy'
    p.oper.type_fft2d = "fft2d.with_pyfftw"

    oper = OperatorsPseudoSpectral3D(params=p)

    field = np.ones(oper.shapeX_loc)

    print(oper.shapeX_loc)
    print(oper.oper2d.shapeX_loc)

    field_fft = oper.fft3d(field)

    assert field_fft.shape == oper.shapeK_loc

    oper.project_perpk3d(field_fft, field_fft, field_fft)

    a2d = np.arange(oper.nx * oper.ny).reshape(oper.oper2d.shapeX_loc)
    a3d = oper.build_invariant_arrayX_from_2d_indices12X(a2d)



"""State of the variables (:mod:`fluidsim.base.sphericalharmo.state`)
=====================================================================

Provides:

.. autoclass:: StateSphericalHarmoSW1L
   :members:
   :private-members:

"""
from fluidsim.base.state import StatePseudoSpectral
from fluidsim.base.setofvariables import SetOfVariables


class StateSphericalHarmoSW1L(StatePseudoSpectral):
    """Contains the state variables and handles the access to fields.

    This is the general class for the pseudo-spectral solvers base on the
    Sperical harmonic transform.

    .. warning::

       We assume incompressibility (div = 0) in this base class!

    """

    @staticmethod
    def _complete_info_solver(info_solver):
        """Static method to complete the ParamContainer info_solver."""

        StatePseudoSpectral._complete_info_solver(info_solver)

        info_solver.classes.State._set_attribs(
            {
                "keys_state_spect": ["rot_sh", "div_sh", "eta_sh"],
                "keys_state_phys": ["ux", "uy", "eta", "rot"],
                "keys_computable": ["div", "q"],
                "keys_phys_needed": [],
                "keys_linear_eigenmodes": [],
            }
        )

    def compute(self, key, SAVE_IN_DICT=True, RAISE_ERROR=True):
        """Compute and return a variable"""
        it = self.sim.time_stepping.it
        if key in self.vars_computed and it == self.it_computed[key]:
            return self.vars_computed[key]

        results = {}

        if key == "ux_sh":
            result = self.oper.sht(self.state_phys.get_var("ux"))
        elif key == "uy_sh":
            result = self.oper.sht(self.state_phys.get_var("uy"))
        elif key == "div":
            result = self.oper.isht(self.state_spect.get_var("div_sh"))
        elif key == "q":
            rot = self.state_phys.get_var("rot")
            eta = self.state_phys.get_var("eta")
            result = rot - self.oper.f_radial * eta
        else:
            to_print = 'Do not know how to compute "' + key + '".'
            if RAISE_ERROR:
                raise ValueError(to_print)
            else:
                print(to_print + "\nreturn an array of zeros.")

                result = self.oper.create_arrayX(value=0.0)

        if SAVE_IN_DICT:
            self.vars_computed[key] = result
            self.it_computed[key] = it

            for key, var in results.items():
                self.vars_computed[key] = var
                self.it_computed[key] = it

        return result

    def statephys_from_statespect(self, state_spect=None, state_phys=None):
        """Compute the state in physical space."""
        if state_spect is None:
            state_spect = self.state_spect

        if state_phys is None:
            state_phys = self.state_phys

        rot_sh = state_spect.get_var("rot_sh")
        div_sh = state_spect.get_var("div_sh")
        eta_sh = state_spect.get_var("eta_sh")

        ux = state_phys.get_var("ux")
        uy = state_phys.get_var("uy")
        eta = state_phys.get_var("eta")
        rot = state_phys.get_var("rot")

        self.oper.vec_from_divrotsh(div_sh, rot_sh, ux, uy)
        self.oper.isht_as_arg(eta_sh, eta)
        self.oper.isht_as_arg(rot_sh, rot)

    def return_statephys_from_statespect(self, state_spect=None):
        """Return the state in physical space as a new object separate from
        ``self.state_phys``.

        """
        state_phys = SetOfVariables(like=self.state_phys)
        self.statephys_from_statespect(state_spect, state_phys)
        return state_phys

    def statespect_from_statephys(self):
        """Compute `state_spect` from `state_phys`."""
        ux = self.state_phys.get_var("ux")
        uy = self.state_phys.get_var("uy")
        eta = self.state_phys.get_var("eta")

        div_sh = self.state_spect.get_var("div_sh")
        rot_sh = self.state_spect.get_var("rot_sh")
        eta_sh = self.state_spect.get_var("eta_sh")
        self.oper.divrotsh_from_vec(ux, uy, div_sh, rot_sh)
        self.oper.sht_as_arg(eta, eta_sh)

    def init_from_uxuyeta(self, ux, uy, eta):
        """Initialize from ``ux``, ``uy`` the velocities and ``eta`` the
        displacement field.

        """
        self.state_phys.set_var("ux", ux)
        self.state_phys.set_var("uy", uy)
        self.state_phys.set_var("eta", eta)

        self.statespect_from_statephys()

        rot_sh = self.state_spect.get_var("rot_sh")
        rot = self.state_phys.get_var("rot")
        self.oper.isht_as_arg(rot_sh, rot)

    def init_from_rotsh(self, rot_sh):
        """Initialize the state from the variable `rot_sh`."""
        # self.oper.dealiasing(rot_sh) -> pointless
        self.state_spect.set_var("rot_sh", rot_sh)
        self.statephys_from_statespect()

    def init_statespect_from(self, **kwargs):
        if len(kwargs) == 1:
            if "rot_sh" in kwargs:
                self.init_from_rotsh(kwargs["rot_sh"])
        else:
            super().init_statespect_from(**kwargs)



"""Test benchmarking (:mod:`fluidsim.util.console.test_bench`)
==============================================================

"""
import unittest
from shutil import rmtree
import sys

from fluiddyn.util import mpi
from fluidsim.util.testing import TestCase, skip_if_no_fluidfft

from fluidsim.util.console.__main__ import run_bench, run_bench_analysis


path_tmp = "/tmp/fluidsim_test_bench"


@skip_if_no_fluidfft
class TestBench(TestCase):
    """Test benchmarking."""

    @classmethod
    def setUpClass(cls):
        if mpi.rank == 0:
            rmtree(path_tmp, ignore_errors=True)

    @classmethod
    def tearDownClass(cls):
        if mpi.rank == 0:
            rmtree(path_tmp, ignore_errors=True)

    def test2d(self):
        """Test launching ns2d benchmarks and plotting results."""

        if mpi.nb_proc > 1:
            type_fft = "fft2d.mpi_with_fftw1d"
        else:
            type_fft = "fft2d.with_pyfftw"

        command = f"fluidsim-bench 24 -d 2 -o {path_tmp} -t {type_fft}"
        sys.argv = command.split()
        run_bench()

        # Can plot only parallel benchmarks
        if mpi.rank == 0 and mpi.nb_proc != 1:

            command = f"fluidsim-bench-analysis 24 -d 2 -i {path_tmp}"
            sys.argv = command.split()
            run_bench_analysis()

        sys.argv = "fluidsim-bench -l -d 2".split()
        run_bench()

        type_fft = "fft2d.mpi_with_fftw1d"

        sys.argv = f"fluidsim-bench -p -d 2 -t {type_fft}".split()
        run_bench()
        sys.argv = f"fluidsim-bench -e -d 2 -t {type_fft}".split()
        run_bench()

    def test3d(self):
        """Test launching ns3d benchmarks and plotting results."""

        if mpi.nb_proc > 1:
            type_fft = "fft3d.mpi_with_fftw1d"
        else:
            type_fft = "fft3d.with_pyfftw"

        command = f"fluidsim-bench 8 -d 3 -o {path_tmp} -t {type_fft}"
        sys.argv = command.split()
        run_bench()

        sys.argv = "fluidsim-bench -l -d 3".split()
        run_bench()

        type_fft = "fft3d.mpi_with_fftw1d"

        sys.argv = f"fluidsim-bench -p -d 3 -t {type_fft}".split()
        run_bench()
        sys.argv = f"fluidsim-bench -e -d 3 -t {type_fft}".split()
        run_bench()


if __name__ == "__main__":
    unittest.main()



"""Forcing (:mod:`fluidsim.solvers.ns3d.forcing`)
=================================================

.. autoclass:: ForcingInternalWavesWatuCoriolis
   :members:

.. autoclass:: ForcingNS3D
   :members:

"""


from pathlib import Path

import numpy as np
import h5netcdf
from scipy.interpolate import interp1d
from math import pi

from transonic import boost, Array

from fluiddyn.util import mpi
from fluidsim.util.frequency_modulation import FrequencyModulatedSignalMaker

from fluidsim.base.forcing.base import ForcingBasePseudoSpectral
from fluidsim.base.forcing.specific import SpecificForcingPseudoSpectralSimple

from .milestone import ForcingMilestone3D


class ForcingInternalWavesWatuCoriolis(SpecificForcingPseudoSpectralSimple):
    """Forcing mimicking an experimental setup in the Coriolis platform.

    The experiments have been carried out within the ERC project `WATU
    <http://nicolas.mordant.free.fr/watu.html>`_. Internal gravity waves are
    forced with two long vertical boards oscillating along their central
    horizontal axes.

    """

    tag = "watu_coriolis"

    @classmethod
    def _complete_params_with_default(cls, params):
        params.forcing.available_types.append(cls.tag)
        params.forcing._set_child(
            cls.tag,
            dict(
                omega_f=0.3,  # rad/s
                delta_omega_f=0.03,  # rad/s
                amplitude=0.05,  # m
                period_forcing=1000,
                approximate_dt=None,
                nb_wave_makers=2,
            ),
        )

    @classmethod
    def _modify_sim_repr_maker(cls, sim_repr_maker):
        p_watu = sim_repr_maker.sim.params.forcing[cls.tag]
        sim_repr_maker.add_parameters(
            {"ampl": p_watu.amplitude, "omegaf": p_watu.omega_f}
        )

    def __init__(self, sim):
        super().__init__(sim)

        self.nb_wave_makers = sim.params.forcing.watu_coriolis.nb_wave_makers
        if self.nb_wave_makers not in (1, 2):
            raise NotImplementedError

        # preparation of a time signal for the forcing
        if mpi.rank != 0:
            self.interpolents = None
        else:
            path_file = Path(sim.output.path_run) / "forcing_watu_coriolis.h5"

            if path_file.exists():
                # load time signals
                with h5netcdf.File(str(path_file), "r") as file:

                    times = file["/times"][...]
                    signals = file["/signals"][...]

            else:
                time_signal_maker = FrequencyModulatedSignalMaker(
                    total_time=sim.params.forcing.watu_coriolis.period_forcing,
                    approximate_dt=sim.params.forcing.watu_coriolis.approximate_dt,
                )

                signals = []
                for _ in range(self.nb_wave_makers):
                    (
                        times,
                        forcing_vs_time,
                    ) = time_signal_maker.create_frequency_modulated_signal(
                        sim.params.forcing.watu_coriolis.omega_f,
                        sim.params.forcing.watu_coriolis.delta_omega_f,
                        sim.params.forcing.watu_coriolis.amplitude,
                    )
                    signals.append(forcing_vs_time)

                signals = np.array(signals)
                # save time signals
                with h5netcdf.File(str(path_file), "w") as file:
                    file.create_variable("/times", ("time",), data=times)
                    file.create_variable(
                        "/signals", ("signal", "time"), data=signals
                    )

            # add one point at the end for periodicity
            signals = np.column_stack((signals, signals[:, 0]))
            times = np.hstack(
                (times, sim.params.forcing.watu_coriolis.period_forcing)
            )

            # interpolation functions
            self.interpolents = [
                interp1d(times, signals[index])
                for index in range(signals.shape[0])
            ]

            # warning : period_forcing / omega_f should be integer
            period_f = 2 * pi / sim.params.forcing.watu_coriolis.omega_f
            if sim.params.forcing.watu_coriolis.period_forcing % period_f > 1e-10:
                mpi.printby0(
                    "WARNING: period_forcing is not a multiple of 2*pi/omega_f."
                    "The forcing velocity signal may not be continuous."
                )

        if mpi.nb_proc > 1:
            self.interpolents = mpi.comm.bcast(self.interpolents, root=0)

        amplitude = sim.params.forcing.watu_coriolis.amplitude
        lx = sim.params.oper.Lx
        ly = sim.params.oper.Ly
        lz = sim.params.oper.Lz
        dx = lx / sim.params.oper.nx

        # calculus of the target velocity components
        width = max(4 * dx, amplitude / 5)

        def step_func(x):
            """Activation function"""
            return 0.5 * (np.tanh(x / width) + 1)

        amplitude_side = amplitude + 0.15
        X, Y, Z = sim.oper.get_XYZ_loc()

        self.maskx = (
            (step_func(-(X - amplitude)) + step_func(X - (lx - amplitude)))
            * step_func(Y - amplitude_side)
            * step_func(-(Y - (ly - amplitude_side)))
        )

        self.masky = (
            (step_func(-(Y - amplitude)) + step_func(Y - (ly - amplitude)))
            * step_func(X - amplitude_side)
            * step_func(-(X - (lx - amplitude_side)))
        )

        z_variation = np.sin(2 * np.pi * Z / lz)
        self.vxtarget = z_variation
        self.vytarget = z_variation

        # calculus of coef_sigma

        # f(t) = f0 * exp(-sigma*t)
        # If we want f(t)/f0 = 10**(-gamma) after n_dt time steps, we have to have:
        # sigma = gamma / (n_dt * dt)

        gamma = 2
        n_dt = 4
        self.coef_sigma = gamma / n_dt

        self.sigma = self.coef_sigma / self.params.time_stepping.deltat_max
        self.period_forcing = self.params.forcing.watu_coriolis.period_forcing

        self._tmp_forcing = sim.oper.create_arrayX()

    def compute_forcing_fft_each_time(self):
        sim = self.sim
        time = sim.time_stepping.t % self.period_forcing
        coef_forcing_time_x = float(self.interpolents[0](time))
        vx = sim.state.state_phys.get_var("vx")
        tmp = self._tmp_forcing
        fx = compute_watu_coriolis_forcing_component(
            self.sigma,
            self.maskx,
            coef_forcing_time_x,
            self.vxtarget,
            vx,
            tmp,
        )
        fx_fft = sim.oper.fft(fx)
        if self.nb_wave_makers == 1:
            return {"vx_fft": fx_fft}
        coef_forcing_time_y = float(self.interpolents[1](time))
        vy = sim.state.state_phys.get_var("vy")
        fy = compute_watu_coriolis_forcing_component(
            self.sigma,
            self.masky,
            coef_forcing_time_y,
            self.vytarget,
            vy,
            tmp,
        )
        return {"vx_fft": fx_fft, "vy_fft": sim.oper.fft(fy)}


A2d = Array[np.float64, "3d"]
A3d = Array[np.float64, "3d"]


@boost
def compute_watu_coriolis_forcing_component(
    sigma: float,
    mask: A2d,
    coef_forcing_time: float,
    target: A3d,
    velocity: A3d,
    out: A3d,
):
    out[:] = sigma * mask * (coef_forcing_time * target - velocity)
    return out


class ForcingNS3D(ForcingBasePseudoSpectral):
    """Main forcing class for the ns3d solver."""

    @staticmethod
    def _complete_info_solver(info_solver, classes=None):
        ForcingBasePseudoSpectral._complete_info_solver(
            info_solver, [ForcingInternalWavesWatuCoriolis, ForcingMilestone3D]
        )



"""Base solver (:mod:`fluidsim.base.solvers.pseudo_spect`)
==========================================================

This module provides two base classes that can be used to define
pseudo-spectral solvers.

.. autoclass:: InfoSolverPseudoSpectral
   :members:
   :private-members:

.. autoclass:: SimulBasePseudoSpectral
   :members:
   :private-members:

.. autoclass:: InfoSolverPseudoSpectral3D
   :members:
   :private-members:

"""

import os

import numpy as np

from fluiddyn.util import mpi

from fluidsim.base.setofvariables import SetOfVariables
from fluidsim.base.solvers.base import SimulBase, InfoSolverBase


class InfoSolverPseudoSpectral(InfoSolverBase):
    """Contain the information on a base pseudo-spectral 2D solver."""

    def _init_root(self):
        """Init. `self` by writting the information on the solver.

        The first-level classes for this base solver are:

        - :class:`fluidsim.base.solvers.pseudo_spect.SimulBasePseudoSpectral`

        - :class:`fluidsim.base.state.StatePseudoSpectral`

        - :class:`fluidsim.base.time_stepping.pseudo_spect.TimeSteppingPseudoSpectral`

        - :class:`fluidsim.operators.operators.OperatorsPseudoSpectral2D`

        """
        super()._init_root()

        self.module_name = "fluidsim.base.solvers.pseudo_spect"
        self.class_name = "SimulBasePseudoSpectral"
        self.short_name = "BasePS"

        self.classes.State.module_name = "fluidsim.base.state"
        self.classes.State.class_name = "StatePseudoSpectral"

        self.classes.TimeStepping.module_name = (
            "fluidsim.base.time_stepping.pseudo_spect"
        )

        self.classes.TimeStepping.class_name = "TimeSteppingPseudoSpectral"

        self.classes.Operators.module_name = "fluidsim.operators.operators"
        if "FLUIDSIM_NO_FLUIDFFT" not in os.environ:
            self.classes.Operators.module_name += "2d"

        self.classes.Operators.class_name = "OperatorsPseudoSpectral2D"

        self.classes.Forcing.class_name = "ForcingBasePseudoSpectral"

        self.classes.Output.class_name = "OutputBasePseudoSpectral"

        self.classes._set_child(
            "Preprocess",
            attribs={
                "module_name": "fluidsim.base.preprocess.pseudo_spect",
                "class_name": "PreprocessPseudoSpectral",
            },
        )


class InfoSolverPseudoSpectral3D(InfoSolverPseudoSpectral):
    """Contain the information on a base pseudo-spectral 3D solver."""

    def _init_root(self):
        """Init. `self` by writting the information on the solver.

        The first-level classes for this base solver are the same as
        for the 2D pseudo-spectral base solver except the class:

        - :class:`fluidsim.operators.operators2d.OperatorsPseudoSpectral3D`

        """

        super()._init_root()

        self.classes.Operators.module_name = "fluidsim.operators.operators3d"
        self.classes.Operators.class_name = "OperatorsPseudoSpectral3D"


class SimulBasePseudoSpectral(SimulBase):
    """Pseudo-spectral base solver."""

    InfoSolver = InfoSolverPseudoSpectral

    @staticmethod
    def _complete_params_with_default(params):
        """Complete the `params` container (static method)."""
        SimulBase._complete_params_with_default(params)

        attribs = {"nu_8": 0.0, "nu_4": 0.0, "nu_m4": 0.0}
        params._set_attribs(attribs)

        params._set_doc(
            params._doc
            + """
nu_8: float

    Hyper-viscous coefficient of order 8. Also used in the method
    compute_freq_diss.

nu_4: float

    Hyper-viscous coefficient of order 4.

nu_m4: float

    Hypo-viscous coefficient of order -4. Hypo-viscosity affect more the large
    scales.

"""
        )

    def compute_freq_diss(self):
        r"""Compute the dissipation frequency.

        Use the `self.params.nu_...` parameters to compute an array
        containing the dissipation frequency as a function of the
        wavenumber.

        .. |p| mathmacro:: \partial

        The governing equations for a pseudo-spectral solver can be
        written as

        .. math:: \p_t S = N(S) - \sigma(k) S,

        where :math:`\sigma` is the frequency associated with the
        linear term.

        In this function, the frequency :math:`\sigma` is split in 2
        parts: the dissipation at small scales and the dissipation at
        large scale (hypo-viscosity, `params.nu_m4`).

        Returns
        -------

        f_d : `numpy.array`
            The dissipation frequency as a function of the wavenumber
            (small sclale part).

        f_d_hypo : `numpy.array`
            The dissipation frequency at large scale (hypo-viscosity)

        """
        if self.params.nu_2 > 0:
            f_d = self.params.nu_2 * self.oper.K2
        else:
            f_d = np.zeros_like(self.oper.K2)

        if self.params.nu_4 > 0.0:
            f_d += self.params.nu_4 * self.oper.K4

        if self.params.nu_8 > 0.0:
            f_d += self.params.nu_8 * self.oper.K8

        if self.params.nu_m4 != 0.0:
            f_d_hypo = self.params.nu_m4 / self.oper.K2_not0 ** 2
            # mode K2 = 0 !
            dim = len(self.oper.shapeK)
            if dim == 2:
                if mpi.rank == 0:
                    f_d_hypo[0, 0] = f_d_hypo[0, 1]
            else:
                if sum(self.oper.seq_indices_first_K) == 0:
                    f_d_hypo[0, 0, 0] = f_d_hypo[0, 0, 1]

        else:
            f_d_hypo = 0.0

        return f_d, f_d_hypo

    def plot_freq_diss(self, direction="x"):
        r"""Plot the dissipation frequencies as functions of k.

        direction : `str`
            Direction of the wavevector to plot with ("x", "y" or "z")

        """

        number = getattr(self.params.oper, "n" + direction)
        length = getattr(self.params.oper, "L" + direction)
        deltak = 2 * np.pi / length
        nk = int(self.params.oper.coef_dealiasing * number // 2)
        ks = deltak * np.arange(nk)

        ks_not0 = ks.copy()
        ks_not0[ks == 0] = np.nan

        fig, ax = self.output.figure_axe()
        ax.set_xlabel(f"$k{direction}$")
        ax.set_ylabel(r"$\omega_\mathrm{diss}$")
        ax.set_title("Dissipation frequencies\n" + self.output.summary_simul)

        f_d_tot = np.zeros_like(ks)

        if self.params.nu_2 > 0:
            f_d_2 = self.params.nu_2 * ks ** 2
            ax.plot(ks, f_d_2, "r", linewidth=2, label=r"$\nu_2$")
            f_d_tot += f_d_2

        if self.params.nu_4 > 0.0:
            f_d_4 = self.params.nu_4 * ks ** 4
            ax.plot(ks, f_d_4, "m", linewidth=2, label=r"$\nu_4$")
            f_d_tot += f_d_4

        if self.params.nu_8 > 0.0:
            f_d_8 = self.params.nu_8 * ks ** 8
            ax.plot(ks, f_d_8, "b", linewidth=2, label=r"$\nu_8$")
            f_d_tot += f_d_8

        if self.params.nu_m4 != 0.0:
            f_d_hypo = self.params.nu_m4 / ks_not0 ** 4
            ax.plot(ks, f_d_hypo, "g", linewidth=2, label=r"$\nu_{-4}$")
            f_d_tot += f_d_hypo

        ax.plot(ks, f_d_tot, "k--", linewidth=2, label="total")

        ax.legend()

    def tendencies_nonlin(self, variables=None, old=None):
        r"""Compute the nonlinear tendencies.

        This function has to be overridden in a child class.

        Returns
        -------

        tendencies_fft : :class:`fluidsim.base.setofvariables.SetOfVariables`
            An array containing only zeros.

        """
        if old is None:
            tendencies = SetOfVariables(like=self.state.state_spect)
        else:
            tendencies = old

        tendencies.initialize(value=0.0)
        return tendencies


Simul = SimulBasePseudoSpectral


if __name__ == "__main__":

    import fluiddyn as fld

    params = Simul.create_default_params()

    params.short_name_type_run = "test"

    nh = 16
    Lh = 2 * np.pi
    params.oper.nx = nh
    params.oper.ny = nh
    params.oper.Lx = Lh
    params.oper.Ly = Lh

    delta_x = params.oper.Lx / params.oper.nx
    params.nu_8 = (
        2.0 * 10e-1 * params.forcing.forcing_rate ** (1.0 / 3) * delta_x ** 8
    )

    params.time_stepping.t_end = 5.0

    params.init_fields.type = "noise"

    params.output.periods_plot.phys_fields = 0.0

    params.output.periods_print.print_stdout = 0.25
    params.output.periods_save.phys_fields = 2.0

    sim = Simul(params)

    sim.output.phys_fields.plot()
    sim.time_stepping.start()
    sim.output.phys_fields.plot()

    fld.show()



"""Run benchmarks (:mod:`fluidsim.util.console.bench`)
======================================================

"""

import gc
from collections import OrderedDict

import numpy as np

from fluiddyn.util import mpi, info
from fluiddyn.io import stdout_redirected

from fluidsim import _is_testing

from ..util import import_module_solver_from_key
from .util import (
    modif_params2d,
    modif_params3d,
    init_parser_base,
    parse_args_dim,
    bench as run_bench,
    tear_down,
    ConsoleError,
)


path_results = "/tmp/fluidsim_bench"
old_print = print
print = mpi.printby0
rank = mpi.rank
nb_proc = mpi.nb_proc
description = "Run benchmarks of FluidSim solvers"


def bench(
    solver,
    dim="2d",
    n0=1024 * 2,
    n1=None,
    n2=None,
    path_dir=None,
    type_fft=None,
    it_end=None,
):
    """Instantiate simulation object and run benchmarks."""

    def _bench(type_fft):
        Simul = solver.Simul
        params = Simul.create_default_params()

        if dim == "2d":
            modif_params2d(params, n0, n1, name_run="bench", type_fft=type_fft)
        elif dim == "3d":
            modif_params3d(
                params, n0, n1, n2, name_run="bench", type_fft=type_fft
            )

        if it_end is not None:
            params.time_stepping.it_end = it_end

        with stdout_redirected():
            sim = Simul(params)

        try:
            run_bench(sim, path_dir)
        except Exception as e:
            if _is_testing:
                raise

            else:
                print(
                    "WARNING: Some error occured while running benchmark"
                    " / saving results!"
                )
                print(e)
        finally:
            tear_down(sim)
            gc.collect()

    if str(type_fft).lower() == "all":
        d = get_opfft(n0, n1, n2, dim, only_dict=True)
        for type_fft, cls in d.items():
            if cls is not None:
                print(type_fft)
                _bench(type_fft)
    else:
        _bench(type_fft)


def get_opfft(n0, n1, n2=None, dim=None, type_fft=None, only_dict=False):
    """Instantiate FFT operator provided by fluidfft."""

    if n2 is None or dim == "2d":
        from fluidfft.fft2d import get_classes_mpi
    elif isinstance(n2, int) or dim == "3d":
        from fluidfft.fft3d import get_classes_mpi

    if mpi.rank == 0:
        d = get_classes_mpi()
    else:
        d = {}

    if mpi.nb_proc > 1:
        d = mpi.comm.bcast(d)

    if only_dict:
        return d

    else:
        if type_fft not in d:
            raise ConsoleError("{} not in {}".format(type_fft, list(d.keys())))

        ClassFFT = d[type_fft]
        if ClassFFT is None:
            raise RuntimeError(f"Class {type_fft} is not available")

        if n2 is None:
            opfft = ClassFFT(n0, n1)
        else:
            opfft = ClassFFT(n0, n1, n2)

        return opfft


def estimate_shapes_weak_scaling(
    n0_max,
    n1_max,
    n2_max=None,
    nproc_min=2,
    nproc_max=None,
    type_fft=None,
    show=False,
):
    """Use this function to get a recommendation of shapeX_seq to initialize the
    solver with to perform weak scaling analysis. The objective is to obtain
    shapeK_loc.

    """
    if nproc_max is None:
        try:
            from os import cpu_count
        except ImportError:
            from multiprocessing import cpu_count

        nproc_max = cpu_count()

    if nproc_min >= nproc_max:
        raise ValueError(
            "Cannot run estimate_shapes_weak_scaling because "
            f"nproc_min >= nproc_max (nproc_max = {nproc_max})"
        )

    if nproc_max % nproc_min != 0:
        raise ValueError(
            "nproc_max % nproc_min != 0 "
            f"(nproc_max={nproc_max}; nproc_min={nproc_min})"
        )

    # Generate a geometric progression for number of processes

    def log(x):
        return int(np.log(x) / np.log(nproc_min))

    num_gp = int(log(nproc_max))
    nproc_gp = np.logspace(1, num_gp, num_gp, base=nproc_min, dtype=int)
    nproc_max = nproc_gp[-1]

    try:
        opfft = get_opfft(n0_max, n1_max, n2_max, type_fft=type_fft)
    except RuntimeError:
        print(f"Cannot create FFT operator {type_fft}")
        return

    shapeX_seq = opfft.get_shapeX_seq()
    shapes = OrderedDict()
    for nproc in nproc_gp:
        divisor = nproc_max // nproc
        if n2_max is None:
            shapes[str(nproc)] = "{} {}".format(
                shapeX_seq[0], shapeX_seq[1] // divisor
            )
        else:
            shapes[str(nproc)] = "{} {}".format(
                shapeX_seq[0], shapeX_seq[1], shapeX_seq[2] // divisor
            )

    if show:
        info._print_heading(["nproc", "shapeX_seq"], case="lower")
        info._print_dict(shapes)

    return shapes


def print_shape_loc(n0, n1, n2=None, type_fft=None):
    """Display the local shape of arrays, shapeX_loc and shapeK_loc.
    Meant to be used with MPI.

    """

    try:
        opfft = get_opfft(n0, n1, n2, type_fft=type_fft)
    except RuntimeError:
        print(f"Cannot create FFT operator {type_fft}")
        return

    shapeX_loc = opfft.get_shapeX_loc()
    shapeK_loc = opfft.get_shapeK_loc()
    print("-" * 8)
    print("type fft = ", type_fft)
    old_print(
        "rank {}: shapeX_loc = {}, shapeK_loc = {}".format(
            rank, shapeX_loc, shapeK_loc
        )
    )


def init_parser(parser):
    """Initialize argument parser for `fluidsim bench`."""

    init_parser_base(parser)
    parser.add_argument("-o", "--output_dir", default=path_results)
    parser.add_argument(
        "-t",
        "--type-fft",
        default=None,
        help=(
            'specify FFT type key (for eg. "fft2d.mpi_with_fftw1d") or "all";'
            "if not specified uses the default FFT method in operators"
        ),
    )
    parser.add_argument(
        "-l",
        "--list-type-fft",
        action="store_true",
        help="list FFT types available for the specified shape or dimension",
    )
    parser.add_argument(
        "-p",
        "--print-shape-loc",
        action="store_true",
        help="mpirun with this option to see how the FFT is initialized",
    )
    parser.add_argument(
        "-e",
        "--estimate-shapes",
        action="store_true",
        help="estimate shapes to plan weak scaling benchmarks",
    )

    parser.add_argument(
        "-it", "--it-end", default=None, type=int, help="Number of iterations"
    )


def run(args):
    """Run `fluidsim bench` command."""

    args = parse_args_dim(args)

    if args.list_type_fft:
        print("FFT classes available for", args.dim.upper())
        d = get_opfft(args.n0, args.n1, args.n2, dim=args.dim, only_dict=True)
        info._print_dict(d)
    elif args.print_shape_loc:
        if args.type_fft is None:
            raise ValueError(
                "Add the fft type, for example -t fft2d.mpi_with_fftw1d"
            )
        print_shape_loc(args.n0, args.n1, args.n2, args.type_fft)
    elif args.estimate_shapes:
        if args.type_fft is None:
            raise ValueError(
                "Add the fft type, for example -t fft2d.mpi_with_fftw1d"
            )
        try:
            estimate_shapes_weak_scaling(
                args.n0, args.n1, args.n2, type_fft=args.type_fft, show=True
            )
        except ValueError as error:
            print(error)

    else:
        # Initialize simulation and run benchmarks
        solver = import_module_solver_from_key(args.solver)
        if args.dim == "3d":
            bench(
                solver,
                args.dim,
                args.n0,
                args.n1,
                args.n2,
                path_dir=args.output_dir,
                type_fft=args.type_fft,
                it_end=args.it_end,
            )
        else:
            bench(
                solver,
                args.dim,
                args.n0,
                args.n1,
                path_dir=args.output_dir,
                type_fft=args.type_fft,
                it_end=args.it_end,
            )



import h5py
import numpy as np

from fluiddyn.util import mpi
from fluidsim.base.output.base import SpecificOutput


class ProbaDensityFunc(SpecificOutput):
    """Handle the saving and plotting of pdf of the turbulent kinetic energy.

    .. todo::

        Rewrite / move this class to as the code is specifically designed for sw1l solvers.

    """

    _tag = "pdf"
    _name_file = _tag + ".h5"

    @staticmethod
    def _complete_params_with_default(params):
        tag = "pdf"

        params.output.periods_save._set_attrib(tag, 0)
        params.output._set_child(tag, attribs={"HAS_TO_PLOT_SAVED": False})

    def __init__(self, output):
        params = output.sim.params
        self.c2 = params.c2
        self.f = params.f
        self.nx = params.oper.nx

        super().__init__(
            output,
            period_save=params.output.periods_save.pdf,
            has_to_plot_saved=params.output.pdf.HAS_TO_PLOT_SAVED,
        )

    def _init_online_plot(self):
        if mpi.rank == 0:
            self.fig, axe = self.output.figure_axe(numfig=5_000_000)
            self.axe = axe
            axe.set_xlabel(r"$\eta$")
            axe.set_ylabel("pdf")
            axe.set_title(r"pdf $\eta$" + "\n" + self.output.summary_simul)

    def _online_plot_saving(self, dict_pdf):
        """online plot on pdf"""
        pdf_eta = dict_pdf["pdf_eta"]
        bin_edges_eta = dict_pdf["bin_edges_eta"]
        self.axe.plot(bin_edges_eta[:-1], pdf_eta, "k")

    def compute(self):
        """compute the values at one time."""
        eta = self.sim.state.state_phys.get_var("eta")
        pdf_eta, bin_edges_eta = self.oper.pdf_normalized(eta)

        ux = self.sim.state.state_phys.get_var("ux")
        uy = self.sim.state.state_phys.get_var("uy")

        uxm = ux.mean()
        uym = uy.mean()

        u_norme = np.sqrt((ux - uxm) ** 2 + (uy - uym) ** 2)
        pdf_u, bin_edges_u = self.oper.pdf_normalized(u_norme)

        dict_pdf = {
            "pdf_eta": pdf_eta,
            "bin_edges_eta": bin_edges_eta,
            "pdf_u": pdf_u,
            "bin_edges_u": bin_edges_u,
        }
        return dict_pdf

    def load(self):
        """load the saved pdf and return a dictionary."""
        with h5py.File(self.path_file, "r") as h5file:

            dset_pdf_eta = h5file["pdf_eta"]
            dset_bin_edges_eta = h5file["bin_edges_eta"]

            pdf_eta = dset_pdf_eta[...]
            bin_edges_eta = dset_bin_edges_eta[...]

            dset_pdf_u = h5file["pdf_u"]
            dset_bin_edges_u = h5file["bin_edges_u"]

            pdf_u = dset_pdf_u[...]
            bin_edges_u = dset_bin_edges_u[...]

            dict_pdf = {
                "pdf_eta": pdf_eta,
                "bin_edges_eta": bin_edges_eta,
                "pdf_u": pdf_u,
                "bin_edges_u": bin_edges_u,
            }
        return dict_pdf

    def plot(self, tmin=0, tmax=1000, delta_t=2):
        """Plot some pdf."""
        x_left_axe = 0.12
        z_bottom_axe = 0.56
        width_axe = 0.85
        height_axe = 0.37
        size_axe = [x_left_axe, z_bottom_axe, width_axe, height_axe]
        fig, ax1 = self.output.figure_axe(size_axe=size_axe)
        ax1.set_xlabel(r"$\eta$")
        ax1.set_ylabel("PDF")
        ax1.set_title("PDF\n" + self.output.summary_simul)
        ax1.set_xscale("linear")
        ax1.set_yscale("linear")

        z_bottom_axe = 0.09
        size_axe[1] = z_bottom_axe
        ax2 = fig.add_axes(size_axe)

        ax2.set_xlabel("$ |\\mathbf{u}-\\langle \\mathbf{u} \\rangle | $")
        ax2.set_ylabel("PDF")
        ax2.set_xscale("linear")
        ax2.set_yscale("linear")

        with h5py.File(self.path_file, "r") as h5file:
            dset_times = h5file["times"]
            times = dset_times[...]
            # nt = len(times)

            dset_pdf_eta = h5file["pdf_eta"]
            dset_bin_edges_eta = h5file["bin_edges_eta"]
            dset_pdf_u = h5file["pdf_u"]
            dset_bin_edges_u = h5file["bin_edges_u"]

            if len(times) > 1:
                delta_t_save = np.mean(times[1:] - times[0:-1])
            else:
                delta_t_save = delta_t

            delta_i_plot = int(np.round(delta_t / delta_t_save))
            if delta_i_plot == 0:
                delta_i_plot = 1
            delta_t = delta_i_plot * delta_t_save

            imin_plot = np.argmin(abs(times - tmin))
            imax_plot = np.argmin(abs(times - tmax))

            tmin_plot = times[imin_plot]
            tmax_plot = times[imax_plot]

            to_print = "plot(tmin={}, tmax={}, delta_t={:.2f})".format(
                tmin, tmax, delta_t
            )
            print(to_print)

            to_print = (
                "plot pdf eta\n"
                "tmin = {0:8.6g} ; tmax = {1:8.6g} ; delta_t = {2:8.6g}"
                "imin = {3:8d} ; imax = {4:8d} ; delta_i = {5:8d}"
            ).format(
                tmin_plot, tmax_plot, delta_t, imin_plot, imax_plot, delta_i_plot
            )
            print(to_print)

            for it in range(imin_plot, imax_plot + 1, delta_i_plot):
                pdf_eta = dset_pdf_eta[it]
                bin_edges_eta = dset_bin_edges_eta[it]

                bin_edges_eta = (bin_edges_eta[:-1] + bin_edges_eta[1:]) / 2
                ax1.plot(bin_edges_eta, pdf_eta, "c", linewidth=1)

                pdf_u = dset_pdf_u[it]
                bin_edges_u = dset_bin_edges_u[it]

                bin_edges_u = (bin_edges_u[:-1] + bin_edges_u[1:]) / 2
                ax2.plot(bin_edges_u, pdf_u, "r", linewidth=1)



"""Initialization of the field (:mod:`fluidsim.solvers.ns2d.strat.init_fields`)
===============================================================================

.. autoclass:: InitFieldsNS2DStrat
   :members:

.. autoclass:: InitFieldsNoiseStrat
   :members:

.. autoclass:: InitFieldsJetStrat
   :members:

.. autoclass:: InitFieldsDipoleStrat
   :members:

"""

import random
import numpy as np

from fluiddyn.util import mpi

from fluidsim.base.init_fields import InitFieldsBase, SpecificInitFields
from fluidsim.solvers.ns2d.init_fields import (
    InitFieldsNoise,
    InitFieldsJet,
    InitFieldsDipole,
)

InitFieldsJetStrat = InitFieldsJet
InitFieldsDipoleStrat = InitFieldsDipole

# It would be nice that rot_fft, ux_fft and uy_fft are attributes of the class
# init_fields.


class InitFieldsNoiseStrat(InitFieldsNoise):
    """
    Class to initialize the fields with noise for the solver ns2d.strat.

    """


class InitFieldsLinearMode(SpecificInitFields):
    """
    Class to initialize the fields with the linear mode
    """

    tag = "linear_mode"

    @classmethod
    def _complete_params_with_default(cls, params):
        """Complete the `params` container (class method)."""
        super()._complete_params_with_default(params)
        p_linear_mode = params.init_fields._set_child(
            cls.tag,
            attribs={
                "i_mode": (8, 8),
                "delta_k_adim": 1,
                "amplitude": 1,
                "eigenmode": "ap_fft",
            },
        )

        p_linear_mode._set_doc(
            """
        i_mode : tuple

            Index of initialized mode (kx, kz).

        delta_k_adim : int

            Size (in modes) of the initialized region.
            Example:
             delta_k_adim = 1 : 1 mode initialized
             delta_k_adim = 2 : 4 modes initialized

        amplitude : float

          Amplitude of the initial linear mode

        eigenmode : str

          Which eigenmode (ap_fft or am_fft)

        """
        )

    def __call__(self):

        params = self.sim.params

        linear_mode = self.put_ones_linear_mode()

        eigenmode = self.sim.params.init_fields.linear_mode.eigenmode

        if eigenmode == "ap_fft":
            print("hello")
            self.sim.state.init_statespect_from(ap_fft=linear_mode)
        elif eigenmode == "am_fft":
            print("Is here?")
            self.sim.state.init_statespect_from(am_fft=linear_mode)
        else:
            raise ValueError("eigenmode should be ap_fft or am_fft.")

    def put_ones_linear_mode(self):
        """Put ones in the linear mode ap_fft or am_fft"""

        oper = self.sim.oper
        params = self.sim.params

        i_mode = params.init_fields.linear_mode.i_mode
        delta_k_adim = params.init_fields.linear_mode.delta_k_adim
        amplitude = params.init_fields.linear_mode.amplitude * params.N ** 2

        # Define linear mode to put energy
        linear_mode = np.zeros(oper.shapeK, dtype="complex")

        # Define grid of indices (sequential or parallel)
        if mpi.nb_proc > 1:
            ikx_mode = np.arange(
                oper.kx_loc[0] / oper.deltakx,
                (oper.kx_loc[-1] + oper.deltakx) / oper.deltakx,
                dtype=int,
            )
            iky_mode = np.arange(0, oper.shapeK[1], dtype=int)

            iKY, iKX = np.meshgrid(iky_mode, ikx_mode)

        else:
            ikx_mode = np.arange(0, oper.shapeK[0])
            iky_mode = np.arange(0, oper.shapeK[1])

            iKX, iKY = np.meshgrid(ikx_mode, iky_mode)

        # Condition to put energy
        cond_x = np.logical_and(
            iKX >= i_mode[0], iKX <= i_mode[0] + delta_k_adim - 1
        )

        cond_y = np.logical_and(
            iKY >= i_mode[1], iKY <= i_mode[1] + delta_k_adim - 1
        )

        COND = np.logical_and(cond_x, cond_y)

        indices_cond = np.argwhere(COND == True)

        for index in indices_cond:
            linear_mode[tuple(index)] = 1.0 + 0j
            # linear_mode[tuple(index)] = random.random() + 0j

        return amplitude * linear_mode


class InitFieldsNS2DStrat(InitFieldsBase):
    """Initialize the state for the solver ns2d.strat."""

    @staticmethod
    def _complete_info_solver(info_solver):
        """Complete the `info_solver` container (static method)."""
        InitFieldsBase._complete_info_solver(
            info_solver,
            classes=[
                InitFieldsNoiseStrat,
                InitFieldsJetStrat,
                InitFieldsDipoleStrat,
                InitFieldsLinearMode,
            ],
        )
